from collections import defaultdict
import functools
import logging
import tempfile
import contextlib

from . import log_utils

LOGGER = logging.getLogger(__name__)
LogTask = functools.partial(log_utils.LogTask, logger=LOGGER)


class LagoAnsible(object):
    """
    A class for handling Ansible related tasks

    Attributes:
        prefix (lago.prefix.Prefix): The prefix that this object wraps
    """

    def __init__(self, prefix):
        """
        Args:
            prefix (lago.prefix.Prefix): A prefix to wrap
        """
        self.prefix = prefix

    def get_inventory_str(self, keys=None):
        """
        Convert a dict generated by ansible.LagoAnsible.get_inventory
        to an INI-like file.

        Args:
            keys (list of str): Path to the keys that will be used to
                create groups.

        Returns:
            str: INI-like Ansible inventory
        """
        inventory = self.get_inventory(keys)
        lines = []
        for name, hosts in inventory.viewitems():
            lines.append('[{name}]'.format(name=name))
            for host in sorted(hosts):
                lines.append(host)

        return '\n'.join(lines)

    def get_inventory(self, keys=None):
        """
        Create an Ansible inventory based on python dicts and lists.
        The returned value is a dict in which every key represents a group
        and every value is a list of entries for that group.

        Args:
            keys (list of str): Path to the keys that will be used to
                create groups.

        Returns:
            dict: dict based Ansible inventory
        """
        inventory = defaultdict(list)
        keys = keys or ['vm-type', 'groups', 'vm-provider']
        vms = self.prefix.get_vms().values()

        for vm in vms:
            entry = self._generate_entry(vm)
            vm_spec = vm.spec
            for key in keys:
                value = self.get_key(key, vm_spec)
                if value is None:
                    continue
                if isinstance(value, list):
                    for sub_value in value:
                        inventory['{}={}'.format(key, sub_value)].append(entry)
                else:
                    inventory['{}={}'.format(key, value)].append(entry)

        return inventory

    def _generate_entry(self, vm):
        """
        Generate host entry for the given VM
        Args:
            vm (lago.plugins.vm.VMPlugin): The VM for which the entry
                should be created for.

        Returns:
            str: An entry for vm
        """
        return \
            '{name} ' \
            'ansible_host={ip} ' \
            'ansible_ssh_private_key_file={key}'.format(
                name=vm.name(),
                ip=vm.ip(),
                key=self.prefix.paths.ssh_id_rsa()
            )

    @staticmethod
    def get_key(key, data_structure):
        """
        Helper method for extracting values from a nested data structure.

        Args:
            key (str): The path to the vales (a series of keys and indexes
                separated by '/')
            data_structure (dict or list): The data structure from which the
                value will be extracted.

        Returns:
            str: The values associated with key
        """
        if key == '/':
            return data_structure

        path = key.split('/')
        # If the path start with '/', remove the empty string from the list
        path[0] or path.pop(0)
        current_value = data_structure
        while path:
            current_key = path.pop(0)
            try:
                current_key = int(current_key)
            except ValueError:
                pass

            try:
                current_value = current_value[current_key]
            except (KeyError, IndexError):
                LOGGER.debug('failed to extract path {}'.format(key))
                return None

        return current_value

    @contextlib.contextmanager
    def get_inventory_temp_file(self, keys=None):
        """
        Context manager which returns the inventory written on a tempfile.
        The tempfile will be deleted as soon as this context manger ends.

        Args:
            keys (list of str): Path to the keys that will be used to
                create groups.

        Yields:
            tempfile.NamedTemporaryFile: Temp file containing the inventory
        """
        temp_file = tempfile.NamedTemporaryFile(mode='r+t')
        inventory = self.get_inventory_str(keys)
        LOGGER.debug(
            'Writing inventory to temp file {} \n{}'.format(
                temp_file.name, inventory
            )
        )
        temp_file.write(inventory)
        temp_file.flush()
        temp_file.seek(0)
        yield temp_file
        temp_file.close()



import array
import fcntl
import functools
import select
import socket
import sys
import termios
import time
import tty
import uuid
import logging

import paramiko

from . import (
    utils,
    log_utils,
)
from .config import config

LOGGER = logging.getLogger(__name__)
LogTask = functools.partial(log_utils.LogTask, logger=LOGGER)
log_task = functools.partial(log_utils.log_task, logger=LOGGER)


def ssh(
    ip_addr,
    command,
    host_name=None,
    data=None,
    show_output=True,
    propagate_fail=True,
    tries=None,
    ssh_key=None,
    username='root',
    password='123456',
):
    host_name = host_name or ip_addr
    client = get_ssh_client(
        ip_addr=ip_addr,
        host_name=host_name,
        propagate_fail=propagate_fail,
        ssh_tries=tries,
        ssh_key=ssh_key,
        username=username,
        password=password,
    )
    transport = client.get_transport()
    channel = transport.open_session()

    joined_command = ' '.join(command)
    command_id = _gen_ssh_command_id()
    LOGGER.debug(
        'Running %s on %s: %s%s',
        command_id,
        host_name,
        joined_command,
        data is not None and (' < "%s"' % data) or '',
    )

    channel.exec_command(joined_command)
    if data is not None:
        channel.send(data)

    channel.shutdown_write()
    return_code, out, err = drain_ssh_channel(
        channel, **(show_output and {} or {
            'stdout': None,
            'stderr': None
        })
    )

    channel.close()
    transport.close()
    client.close()

    LOGGER.debug(
        'Command %s on %s returned with %d',
        command_id,
        host_name,
        return_code,
    )

    if out:
        LOGGER.debug(
            'Command %s on %s output:\n %s',
            command_id,
            host_name,
            out,
        )
    if err:
        LOGGER.debug(
            'Command %s on %s  errors:\n %s',
            command_id,
            host_name,
            err,
        )
    return utils.CommandStatus(return_code, out, err)


def wait_for_ssh(
    ip_addr,
    host_name=None,
    connect_timeout=600,  # 10 minutes
    ssh_key=None,
    username='root',
    password='123456',
):
    host_name = host_name or ip_addr
    start_time = time.time()
    while (time.time() - start_time) < connect_timeout:
        try:
            ret, _, _ = ssh(
                ip_addr=ip_addr,
                host_name=host_name,
                command=['true'],
                tries=1,
                propagate_fail=False,
                ssh_key=ssh_key,
                username=username,
                password=password,
            )
        except Exception as err:
            ret = -1
            sys.exc_clear()
            LOGGER.debug(
                'Got exception while sshing to %s: %s',
                host_name,
                err,
            )

        if ret == 0:
            break

        time.sleep(1)
    else:
        # Try one last time, using the ssh default timeout values, as we
        # already waited for boot_time_sec for sure
        ret, _, _ = ssh(
            ip_addr=ip_addr,
            host_name=host_name,
            command=['true'],
            ssh_key=ssh_key,
            username=username,
            password=password,
        )
        if ret != 0:
            raise RuntimeError(
                'Failed to connect remote shell to %s',
                host_name,
            )

    LOGGER.debug('Wait succeeded for ssh to %s', host_name)


def ssh_script(
    ip_addr,
    path,
    host_name=None,
    show_output=True,
    ssh_key=None,
    username='root',
    password='123456',
):
    host_name = host_name or ip_addr
    LOGGER.debug('Running %s on host %s', path, host_name)
    with open(path) as script_fd:
        return ssh(
            ip_addr=ip_addr,
            host_name=host_name,
            command=['bash', '-s'],
            data=script_fd.read(),
            show_output=show_output,
            ssh_key=ssh_key,
            username=username,
            password=password,
        )


def interactive_ssh(
    ip_addr,
    command=None,
    host_name=None,
    ssh_key=None,
    username='root',
    password='123456',
):
    if command is None:
        command = ['bash']

    client = get_ssh_client(
        ip_addr=ip_addr,
        host_name=host_name,
        ssh_key=ssh_key,
        username=username,
        password=password,
    )
    transport = client.get_transport()
    channel = transport.open_session()
    try:
        return interactive_ssh_channel(channel, ' '.join(command))
    finally:
        channel.close()
        transport.close()
        client.close()


def drain_ssh_channel(chan, stdin=None, stdout=sys.stdout, stderr=sys.stderr):
    chan.settimeout(0)
    out_queue = []
    out_all = []
    err_queue = []
    err_all = []

    try:
        stdout_is_tty = stdout.isatty()
        tty_w = tty_h = -1
    except AttributeError:
        stdout_is_tty = False

    done = False
    while not done:
        if stdout_is_tty:
            arr = array.array('h', range(4))
            if not fcntl.ioctl(stdout.fileno(), termios.TIOCGWINSZ, arr):
                if tty_h != arr[0] or tty_w != arr[1]:
                    tty_h, tty_w = arr[:2]
                    chan.resize_pty(width=tty_w, height=tty_h)

        read_streams = []
        if not chan.closed:
            read_streams.append(chan)

            if stdin and not stdin.closed:
                read_streams.append(stdin)

        write_streams = []
        if stdout and out_queue:
            write_streams.append(stdout)
        if stderr and err_queue:
            write_streams.append(stderr)

        read, write, _ = select.select(
            read_streams,
            write_streams,
            [],
            0.1,
        )

        if stdin in read:
            chunk = utils.read_nonblocking(stdin)
            if chunk:
                chan.send(chunk)
            else:
                chan.shutdown_write()

        try:
            if chan.recv_ready():
                chunk = chan.recv(1024)
                if stdout:
                    out_queue.append(chunk)
                out_all.append(chunk)

            if chan.recv_stderr_ready():
                chunk = chan.recv_stderr(1024)
                if stderr:
                    err_queue.append(chunk)
                err_all.append(chunk)
        except socket.error:
            pass

        if stdout in write:
            stdout.write(out_queue.pop(0))
            stdout.flush()
        if stderr in write:
            stderr.write(err_queue.pop(0))
            stderr.flush()

        if chan.closed and not out_queue and not err_queue:
            done = True

    return (chan.exit_status, ''.join(out_all), ''.join(err_all))


def interactive_ssh_channel(chan, command=None, stdin=sys.stdin):
    try:
        stdin_is_tty = stdin.isatty()
    except Exception:
        stdin_is_tty = False

    if stdin_is_tty:
        oldtty = termios.tcgetattr(stdin)
        chan.get_pty()

    if command is not None:
        chan.exec_command(command)

    try:
        if stdin_is_tty:
            tty.setraw(stdin.fileno())
            tty.setcbreak(stdin.fileno())
        return utils.CommandStatus(*drain_ssh_channel(chan, stdin))
    finally:
        if stdin_is_tty:
            termios.tcsetattr(stdin, termios.TCSADRAIN, oldtty)


def _gen_ssh_command_id():
    return uuid.uuid1().hex[:8]


def get_ssh_client(
    ip_addr,
    ssh_key=None,
    host_name=None,
    ssh_tries=None,
    propagate_fail=True,
    username='root',
    password='123456',
):
    host_name = host_name or ip_addr
    with LogTask(
        'Get ssh client for %s' % host_name,
        level='debug',
        propagate_fail=propagate_fail,
    ):
        ssh_timeout = int(config.get('ssh_timeout'))
        if ssh_tries is None:
            ssh_tries = int(config.get('ssh_tries', 10))

        start_time = time.time()
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy(), )
        while ssh_tries > 0:
            try:
                client.connect(
                    ip_addr,
                    username=username,
                    password=password,
                    key_filename=ssh_key,
                    timeout=ssh_timeout,
                )
                break
            except (socket.error, socket.timeout) as err:
                LOGGER.debug(
                    'Socket error connecting to %s: %s',
                    host_name,
                    err,
                )
            except paramiko.ssh_exception.SSHException as err:
                LOGGER.debug(
                    'SSH error connecting to %s: %s',
                    host_name,
                    err,
                )
            except EOFError as err:
                LOGGER.debug('EOFError connecting to %s: %s', host_name, err)
            ssh_tries -= 1
            LOGGER.debug(
                'Still got %d tries for %s',
                ssh_tries,
                host_name,
            )
            time.sleep(1)
        else:
            end_time = time.time()
            raise RuntimeError(
                'Timed out (in %d s) trying to ssh to %s' %
                (end_time - start_time, host_name)
            )

    return client



#
# Copyright 2014 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
"""
About CLIPlugins

A CLIPlugin is a subcommand of the lagocli command, it's ment to group
actions together in a logical sense, for example grouping all the actions
done to templates.

To create a new subcommand for testenvcli you just have to subclass the
CLIPlugin abstract class and declare it in the setuptools as an entry_point,
see this module's setup.py/setup.cfg for an example::

    class NoopCLIplugin(CLIPlugin):
        init_args = {
            'help': 'dummy help string',
        }

        def populate_parser(self, parser):
            parser.addArgument('--dummy-flag', action='store_true')

        def do_run(self, args):
            if args.dummy_flag:
                print "Dummy flag passed to noop subcommand!"
            else:
                print "Dummy flag not passed to noop subcommand!"


You can also use decorators instead, an equivalent is::

    @cli_plugin_add_argument('--dummy-flag', action='store_true')
    @cli_plugin(help='dummy help string')
    def my_fancy_plugin_func(dummy_flag, **kwargs):
        if dummy_flag:
            print "Dummy flag passed to noop subcommand!"
        else:
            print "Dummy flag not passed to noop subcommand!"

Or::

    @cli_plugin_add_argument('--dummy-flag', action='store_true')
    def my_fancy_plugin_func(dummy_flag, **kwargs):
        "dummy help string"
        if dummy_flag:
            print "Dummy flag passed to noop subcommand!"
        else:
            print "Dummy flag not passed to noop subcommand!"


Then you will need to add an entry_points section in your setup.py like::

    setup(
        ...
        entry_points={
            'lago.plugins.cli': [
                'noop=noop_module:my_fancy_plugin_func',
            ],
        }
        ...
    )


Or in your setup.cfg like::

    [entry_points]
    lago.plugins.cli =
        noop=noop_module:my_fancy_plugin_func



Any of those will add a new subcommand to the lagocli command that can be run
as::

    $ lagocli noop
    Dummy flag not passed to noop subcommand!

TODO: Allow per-plugin namespacing to get rid of the `**kwargs` parameter
"""

import functools
from abc import (
    abstractmethod,
    abstractproperty,
    ABCMeta,
)

from . import Plugin


class CLIPlugin(Plugin):
    __metaclass__ = ABCMeta

    def __init__(self):
        pass

    @abstractproperty
    def init_args(self):
        """
        Dictionary with the argument to initialize the cli parser (for
        example, the help argument)
        """

    @abstractmethod
    def populate_parser(self, parser):
        """
        Add any required arguments to the parser

        Args:
            parser (ArgumentParser): parser to add the arguments to

        Returns:
            None
        """

    @abstractmethod
    def do_run(self, args):
        """
        Execute any actions given the arguments

        Args:
            args (Namespace): with the arguments

        Returns:
            None
        """


class CLIPluginFuncWrapper(CLIPlugin):
    """
    Special class to handle decorated cli plugins, take into account that the
    decorated functions have some limitations on what arguments can they
    define actually, if you need something complicated, used the abstract class
    :class:`CLIPlugin` instead.

    Keep in mind that right now the decorated function must use `**kwargs` as
    param, as it will be passed all the members of the parser, not just
    whatever it defined
    """

    def __init__(self, do_run=None, init_args=None):
        self._init_args = init_args or {}
        self._parser_args = []
        self._do_run = do_run
        if do_run:
            self.set_help()

    @property
    def init_args(self):
        return self._init_args

    def set_help(self, help=None):
        self._init_args['help'] = (
            help if help is not None else self._do_run.__doc__
        )

    def set_init_args(self, init_args):
        self._init_args.update(init_args)

    def add_argument(self, *argument_args, **argument_kwargs):
        self._parser_args.append((argument_args, argument_kwargs))

    def populate_parser(self, parser):
        for argument_args, argument_kwargs in self._parser_args:
            parser.add_argument(*argument_args, **argument_kwargs)

    def do_run(self, args):
        self._do_run(**vars(args))

    def __call__(self, *args, **kwargs):
        """
        Keep the original function interface, so it can be used elsewhere
        """
        return self._do_run(*args, **kwargs)


def cli_plugin_add_argument(*args, **kwargs):
    """
    Decorator generator that adds an argument to the cli plugin based on the
    decorated function

    Args:
        *args: Any args to be passed to
            :func:`argparse.ArgumentParser.add_argument`
        *kwargs: Any keyword args to be passed to
            :func:`argparse.ArgumentParser.add_argument`


    Returns:
        function: Decorator that builds or extends the cliplugin for the
            decorated function, adding the given argument definition

    Examples:
        >>> @cli_plugin_add_argument('-m', '--mogambo', action='store_true')
        ... def test(**kwargs):
        ...     print 'test'
        ...
        >>> print test.__class__
        <class 'cli.CLIPluginFuncWrapper'>
        >>> print test._parser_args
        [(('-m', '--mogambo'), {'action': 'store_true'})]

        >>> @cli_plugin_add_argument('-m', '--mogambo', action='store_true')
        ... @cli_plugin_add_argument('-b', '--bogabmo', action='store_false')
        ... @cli_plugin
        ... def test(**kwargs):
        ...     print 'test'
        ...
        >>> print test.__class__
        <class 'cli.CLIPluginFuncWrapper'>
        >>> print test._parser_args # doctest: +NORMALIZE_WHITESPACE
        [(('-b', '--bogabmo'), {'action': 'store_false'}),
         (('-m', '--mogambo'), {'action': 'store_true'})]

    """

    def decorator(func):
        if not isinstance(func, CLIPluginFuncWrapper):
            func = CLIPluginFuncWrapper(do_run=func)

        func.add_argument(*args, **kwargs)
        return func

    return decorator


def cli_plugin_add_help(help):
    """
    Decorator generator that adds the cli help to the cli plugin based on the
    decorated function

    Args:
        help (str): help string for the cli plugin

    Returns:
        function: Decorator that builds or extends the cliplugin for the
            decorated function, setting the given help
    Examples:
        >>> @cli_plugin_add_help('my help string')
        ... def test(**kwargs):
        ...     print 'test'
        ...
        >>> print test.__class__
        <class 'cli.CLIPluginFuncWrapper'>
        >>> print test.help
        my help string

        >>> @cli_plugin_add_help('my help string')
        ... @cli_plugin()
        ... def test(**kwargs):
        ...     print 'test'
        >>> print test.__class__
        <class 'cli.CLIPluginFuncWrapper'>
        >>> print test.help
        my help string
    """

    def decorator(func):

        if not isinstance(func, CLIPluginFuncWrapper):
            func = CLIPluginFuncWrapper(do_run=func)

        func.set_help(help)
        return func

    return decorator


def cli_plugin(func=None, **kwargs):
    """
    Decorator that wraps the given function in a :class:`CLIPlugin`

    Args:
        func (callable): function/class to decorate
        **kwargs: Any other arg to use when initializing the parser (like help,
            or prefix_chars)

    Returns:
        CLIPlugin: cli plugin that handles that method

    Notes:
        It can be used as a decorator or as a decorator generator, if used as a
        decorator generator don't pass any parameters

    Examples:
        >>> @cli_plugin
        ... def test(**kwargs):
        ...     print 'test'
        ...
        >>> print test.__class__
        <class 'cli.CLIPluginFuncWrapper'>

        >>> @cli_plugin()
        ... def test(**kwargs):
        ...     print 'test'
        >>> print test.__class__
        <class 'cli.CLIPluginFuncWrapper'>

        >>> @cli_plugin(help='dummy help')
        ... def test(**kwargs):
        ...     print 'test'
        >>> print test.__class__
        <class 'cli.CLIPluginFuncWrapper'>
        >>> print test.init_args['help']
        'dummy help'

    """
    # this allows calling this function as a decorator generator
    if func is None:
        return functools.partial(cli_plugin, **kwargs)

    if not isinstance(func, CLIPluginFuncWrapper):
        func = CLIPluginFuncWrapper(do_run=func, init_args=kwargs)
    else:
        func.set_init_args(kwargs)

    return func



#
# Copyright 2014 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
import Queue
import collections
import datetime
import fcntl
import functools
import json
import logging
import os
import shutil
import signal
import subprocess
import sys
import threading
import textwrap
import time
import yaml
import pkg_resources
from io import StringIO
import lockfile
import argparse
import configparser
import uuid as uuid_m
from . import constants
from .log_utils import (LogTask, setup_prefix_logging)
import hashlib

LOGGER = logging.getLogger(__name__)


class TimerException(Exception):
    """
    Exception to throw when a timeout is reached
    """
    pass


def _ret_via_queue(func, queue):
    try:
        queue.put({'return': func()})
    except Exception:
        LOGGER.exception('Error while running thread')
        queue.put({'exception': sys.exc_info()})


def func_vector(target, args_sequence):
    return [functools.partial(target, *args) for args in args_sequence]


class VectorThread:
    def __init__(self, targets):
        self.targets = targets
        self.results = None

    def start_all(self):
        self.thread_handles = []
        for target in self.targets:
            q = Queue.Queue()
            t = threading.Thread(target=_ret_via_queue, args=(target, q))
            self.thread_handles.append((t, q))
            t.start()

    def join_all(self, raise_exceptions=True):
        if self.results:
            return self.results

        for t, q in self.thread_handles:
            t.join()

        self.results = map(lambda (t, q): q.get(), self.thread_handles)
        if raise_exceptions:
            for result in self.results:
                if 'exception' in result:
                    exc_info = result['exception']
                    raise exc_info[1], None, exc_info[2]
        return map(lambda x: x.get('return', None), self.results)


def invoke_in_parallel(func, *args_sequences):
    vt = VectorThread(func_vector(func, zip(*args_sequences)))
    vt.start_all()
    return vt.join_all()


def invoke_different_funcs_in_parallel(*funcs):
    vt = VectorThread(funcs)
    vt.start_all()
    return vt.join_all()


_CommandStatus = collections.namedtuple(
    'CommandStatus', ('code', 'out', 'err')
)


class CommandStatus(_CommandStatus):
    def __nonzero__(self):
        return self.code


def _run_command(
    command,
    input_data=None,
    stdin=None,
    out_pipe=subprocess.PIPE,
    err_pipe=subprocess.PIPE,
    env=None,
    uuid=None,
    **kwargs
):
    """
    Runs a command

    Args:
        command(list of str): args of the command to execute, including the
            command itself as command[0] as `['ls', '-l']`
        input_data(str): If passed, will feed that data to the subprocess
            through stdin
        out_pipe(int or file): File descriptor as passed to
            :ref:subprocess.Popen to use as stdout
        stdin(int or file): File descriptor as passed to
            :ref:subprocess.Popen to use as stdin
        err_pipe(int or file): File descriptor as passed to
            :ref:subprocess.Popen to use as stderr
        env(dict of str:str): If set, will use the given dict as env for the
            subprocess
        uuid(uuid): If set the command will be logged with the given uuid
            converted to string, otherwise, a uuid v4 will be generated.
        **kwargs: Any other keyword args passed will be passed to the
            :ref:subprocess.Popen call

    Returns:
        lago.utils.CommandStatus: result of the interactive execution
    """

    # add libexec to PATH if needed
    if uuid is None:
        uuid = uuid_m.uuid4()

    if constants.LIBEXEC_DIR not in os.environ['PATH'].split(':'):
        os.environ['PATH'
                   ] = '%s:%s' % (constants.LIBEXEC_DIR, os.environ['PATH'])

    if input_data and not stdin:
        kwargs['stdin'] = subprocess.PIPE
    elif stdin:
        kwargs['stdin'] = stdin

    if env is None:
        env = os.environ.copy()
    else:
        env['PATH'] = ':'.join(
            list(
                set(
                    env.get('PATH', '').split(':') + os.environ['PATH']
                    .split(':')
                ),
            ),
        )

    popen = subprocess.Popen(
        ' '.join('"%s"' % arg for arg in command),
        stdout=out_pipe,
        stderr=err_pipe,
        shell=True,
        env=env,
        **kwargs
    )
    out, err = popen.communicate(input_data)
    LOGGER.debug(
        '%s: command exit with return code: %d', str(uuid), popen.returncode
    )
    if out:
        LOGGER.debug('%s: command stdout: %s', str(uuid), out)
    if err:
        LOGGER.debug('%s: command stderr: %s', str(uuid), err)
    return CommandStatus(popen.returncode, out, err)


def run_command(
    command,
    input_data=None,
    out_pipe=subprocess.PIPE,
    err_pipe=subprocess.PIPE,
    env=None,
    **kwargs
):
    """
    Runs a command non-interactively

    Args:
        command(list of str): args of the command to execute, including the
            command itself as command[0] as `['ls', '-l']`
        input_data(str): If passed, will feed that data to the subprocess
            through stdin
        out_pipe(int or file): File descriptor as passed to
            :ref:subprocess.Popen to use as stdout
        err_pipe(int or file): File descriptor as passed to
            :ref:subprocess.Popen to use as stderr
        env(dict of str:str): If set, will use the given dict as env for the
            subprocess
        **kwargs: Any other keyword args passed will be passed to the
            :ref:subprocess.Popen call

    Returns:
        lago.utils.CommandStatus: result of the interactive execution
    """
    if env is None:
        env = os.environ.copy()

    with LogTask(
        'Run command: %s' % ' '.join('"%s"' % arg for arg in command),
        logger=LOGGER,
        level='debug',
    ) as task:
        command_result = _run_command(
            command=command,
            input_data=input_data,
            out_pipe=out_pipe,
            err_pipe=err_pipe,
            env=env,
            uuid=task.uuid,
            **kwargs
        )
        return command_result


def run_interactive_command(command, env=None, **kwargs):
    """
    Runs a command interactively, reusing the current stdin, stdout and stderr

    Args:
        command(list of str): args of the command to execute, including the
            command itself as command[0] as `['ls', '-l']`
        env(dict of str:str): If set, will use the given dict as env for the
            subprocess
        **kwargs: Any other keyword args passed will be passed to the
            :ref:subprocess.Popen call

    Returns:
        lago.utils.CommandStatus: result of the interactive execution
    """
    command_result = _run_command(
        command=command,
        out_pipe=sys.stdout,
        err_pipe=sys.stderr,
        stdin=sys.stdin,
        env=env,
        **kwargs
    )
    return command_result


def service_is_enabled(name):
    ret, out, _ = run_command(['systemctl', 'is-enabled', name])
    if ret == 0 and out.strip() == 'enabled':
        return True
    return False


# Copied from VDSM: lib/vdsm/utils.py
class RollbackContext(object):
    '''
    A context manager for recording and playing rollback.
    The first exception will be remembered and re-raised after rollback

    Sample usage:
    > with RollbackContext() as rollback:
    >     step1()
    >     rollback.prependDefer(lambda: undo step1)
    >     def undoStep2(arg): pass
    >     step2()
    >     rollback.prependDefer(undoStep2, arg)

    More examples see tests/utilsTests.py @ vdsm code
    '''

    def __init__(self, *args):
        self._finally = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        If this function doesn't return True (or raises a different
        exception), python re-raises the original exception once this
        function is finished.
        """
        undoExcInfo = None
        for undo, args, kwargs in self._finally:
            try:
                undo(*args, **kwargs)
            except Exception:
                # keep the earliest exception info
                if undoExcInfo is None:
                    undoExcInfo = sys.exc_info()

        if exc_type is None and undoExcInfo is not None:
            raise undoExcInfo[0], undoExcInfo[1], undoExcInfo[2]

    def defer(self, func, *args, **kwargs):
        self._finally.append((func, args, kwargs))

    def prependDefer(self, func, *args, **kwargs):
        self._finally.insert(0, (func, args, kwargs))

    def clear(self):
        self._finally = []


class EggTimer:
    def __init__(self, timeout):
        self.timeout = timeout

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, *_):
        pass

    def elapsed(self):
        return (time.time() - self.start_time) > self.timeout


class ExceptionTimer(object):
    def __init__(self, timeout):
        self.timeout = timeout or 0

    def __enter__(self):
        self.start()

    def __exit__(self, *_):
        self.stop()

    def start(self):
        def raise_timeout(*_):
            raise TimerException('Passed %d seconds' % self.timeout)

        signal.signal(signal.SIGALRM, raise_timeout)
        signal.alarm(self.timeout)

    def stop(self):
        signal.alarm(0)


class LockFile(object):
    """
    Context manager that creates a lock around a directory, with optional
    timeout in the acquire operation

    Args:
        path(str): path to the dir to lock
        timeout(int): timeout in seconds to wait while acquiring the lock
        **kwargs(dict): Any other param to pass to `lockfile.LockFile`
    """

    def __init__(self, path, timeout=None, **kwargs):
        self.path = path
        self.timeout = timeout or 0
        self.lock = lockfile.LockFile(path=path, **kwargs)

    def __enter__(self):
        """
        Start the lock with timeout if needed in the acquire operation

        Raises:
            TimerException: if the timeout is reached before acquiring the lock
        """
        try:
            with ExceptionTimer(timeout=self.timeout):
                LOGGER.debug('Acquiring lock for {}'.format(self.path))
                self.lock.acquire()
                LOGGER.debug('Holding the lock for {}'.format(self.path))
        except TimerException:
            raise TimerException(
                'Unable to acquire lock for %s in %s secs',
                self.path,
                self.timeout,
            )

    def __exit__(self, *_):
        if self.lock.is_locked():
            LOGGER.debug('Trying to release lock for {}'.format(self.path))
            self.lock.release()
            LOGGER.debug('Lock for {} was released'.format(self.path))


def read_nonblocking(file_descriptor):
    oldfl = fcntl.fcntl(file_descriptor.fileno(), fcntl.F_GETFL)
    try:
        fcntl.fcntl(
            file_descriptor.fileno(),
            fcntl.F_SETFL,
            oldfl | os.O_NONBLOCK,
        )
        return file_descriptor.read()
    finally:
        fcntl.fcntl(file_descriptor.fileno(), fcntl.F_SETFL, oldfl)


def json_dump(obj, f):
    return json.dump(obj, f, indent=4)


def deepcopy(original_obj):
    """
    Creates a deep copy of an object with no crossed referenced lists or dicts,
    useful when loading from yaml as anchors generate those cross-referenced
    dicts and lists

    Args:
        original_obj(object): Object to deep copy

    Return:
        object: deep copy of the object
    """
    if isinstance(original_obj, list):
        return list(deepcopy(item) for item in original_obj)
    elif isinstance(original_obj, dict):
        return dict((key, deepcopy(val)) for key, val in original_obj.items())
    else:
        return original_obj


def load_virt_stream(virt_fd):
    """
    Loads the given conf stream into a dict, trying different formats if
    needed

    Args:
        virt_fd (str): file like objcect with the virt config to load

    Returns:
        dict: Loaded virt config
    """
    try:
        virt_conf = json.load(virt_fd)
    except ValueError:
        virt_fd.seek(0)
        virt_conf = yaml.load(virt_fd)

    return deepcopy(virt_conf)


def in_prefix(prefix_class, workdir_class):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            prefix_path = kwargs.get('prefix_path', None)
            workdir_path = kwargs.get('workdir_path', None)
            if (
                prefix_path is not None or (
                    prefix_class.is_prefix(os.curdir) and workdir_path is None
                )
            ):
                LOGGER.debug('Looking for a prefix')
                prefix_path = os.path.realpath(
                    prefix_class.resolve_prefix_path(prefix_path)
                )
                prefix = prefix_class(prefix_path)
                kwargs['parent_workdir'] = None

            else:
                LOGGER.debug('Looking for a workdir')
                if workdir_path is None:
                    workdir_path = 'auto'

                workdir_path = workdir_class.resolve_workdir_path(workdir_path)
                workdir = workdir_class(path=workdir_path)
                kwargs['parent_workdir'] = workdir
                if kwargs.get('all_envs', False):
                    prefix = workdir
                else:
                    prefix_name = kwargs.get('prefix_name', 'current')
                    prefix = workdir.get_prefix(prefix_name)
                    kwargs['prefix_name'] = prefix_name

                prefix_path = os.path.realpath(
                    os.path.join(workdir_path, prefix_name)
                )

            kwargs['prefix'] = prefix
            os.environ['LAGO_PREFIX_PATH'] = prefix_path or ''
            os.environ['LAGO_WORKDIR_PATH'] = workdir_path or ''
            return func(*args, **kwargs)

        return wrapper

    return decorator


def with_logging(func):
    @functools.wraps(func)
    def wrapper(prefix, *args, **kwargs):
        setup_prefix_logging(prefix.paths.logs())
        return func(*args, prefix=prefix, **kwargs)

    return wrapper


def add_timestamp_suffix(base_string):
    return datetime.datetime.fromtimestamp(
        time.time()
    ).strftime(base_string + '.%Y-%m-%d_%H:%M:%S')


def rotate_dir(base_dir):
    shutil.move(base_dir, add_timestamp_suffix(base_dir))


def ipv4_to_mac(ip):
    # Mac addrs of domains are 54:52:xx:xx:xx:xx where the last 4 octets are
    # the hex repr of the IP address)
    mac_addr_pieces = [0x54, 0x52] + [int(y) for y in ip.split('.')]
    return ':'.join([('%02x' % x) for x in mac_addr_pieces])


def argparse_to_ini(parser, root_section='lago', incl_unset=False):
    subparsers_actions = [
        action for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)
    ]

    root_actions = [
        action for action in parser._actions
        if not isinstance(action, argparse._SubParsersAction)
    ]

    cp = configparser.ConfigParser(allow_no_value=True)

    _add_subparser_to_cp(cp, root_section, root_actions, incl_unset)
    for subparsers_action in subparsers_actions:
        for choice, subparser in subparsers_action.choices.items():
            _add_subparser_to_cp(cp, choice, subparser._actions, incl_unset)

    header = '# Lago configuration file, generated: {0}'.format(
        time.strftime("%c")
    )
    with StringIO() as ini_str:
        cp.write(ini_str)
        return '\n'.join([header, ini_str.getvalue()])


def _add_subparser_to_cp(cp, section, actions, incl_unset):
    cp.add_section(section)
    print_actions = (
        action for action in actions
        if (action.default and action.default != '==SUPPRESS==') or
        (action.default is None and incl_unset)
    )
    for action in print_actions:
        var = str(action.dest)
        if action.default is None:
            var = '#{0}'.format(var)
        if action.help:
            for line in textwrap.wrap(action.help, width=70):
                cp.set(section, '# {0}'.format(line))
        cp.set(section, var, str(action.default))
    if len(cp.items(section)) == 0:
        cp.remove_section(section)


def run_command_with_validation(
    cmd, fail_on_error=True, msg='An error has occurred'
):
    result = run_command(cmd)
    if result and fail_on_error:
        raise RuntimeError('{}\n{}'.format(msg, result.err))

    return result


def get_qemu_info(path, backing_chain=False, fail_on_error=True):
    """
    Get info on a given qemu disk

    Args:
        path(str): Path to the required disk
        backing_chain(boo): if true, include also info about
        the image predecessors.
    Return:
        object: if backing_chain == True then a list of dicts else a dict
    """

    cmd = ['qemu-img', 'info', '--output=json', path]

    if backing_chain:
        cmd.insert(-1, '--backing-chain')

    result = run_command_with_validation(
        cmd, fail_on_error, msg='Failed to get info for {}'.format(path)
    )

    return json.loads(result.out)


def qemu_rebase(target, backing_file, safe=True, fail_on_error=True):
    """
    changes the backing file of 'source' to 'backing_file'
    If backing_file is specified as "" (the empty string),
    then the image is rebased onto no backing file
    (i.e. it will exist independently of any backing file).
    (Taken from qemu-img man page)

    Args:
        target(str): Path to the source disk
        backing_file(str): path to the base disk
        safe(bool): if false, allow unsafe rebase
         (check qemu-img docs for more info)
    """
    cmd = ['qemu-img', 'rebase', '-b', backing_file, target]
    if not safe:
        cmd.insert(2, '-u')

    return run_command_with_validation(
        cmd,
        fail_on_error,
        msg='Failed to rebase {target} onto {backing_file}'.format(
            target=target, backing_file=backing_file
        )
    )


def compress(input_file, block_size, fail_on_error=True):
    cmd = [
        'xz', '--compress', '--keep', '--threads=0', '--best', '--force',
        '--verbose', '--block-size={}'.format(block_size), input_file
    ]
    return run_command_with_validation(
        cmd, fail_on_error, msg='Failed to compress {}'.format(input_file)
    )


def cp(input_file, output_file, fail_on_error=True):
    if not os.path.basename(output_file):
        output_file = os.path.join(output_file, os.path.basename(input_file))

    cmd = ['cp', input_file, output_file]
    return run_command_with_validation(
        cmd,
        fail_on_error,
        msg='Failed to copy {} to {}'.format(input_file, output_file)
    )


def sparse(input_file, input_format, fail_on_error=True):
    cmd = [
        'virt-sparsify',
        '-q',
        '-v',
        '--format',
        input_format,
        '--in-place',
        input_file,
    ]
    return run_command_with_validation(
        cmd, fail_on_error, msg='Failed to sparse {}'.format(input_file)
    )


def get_hash(file_path, checksum='sha1'):
    """
    Generate a hash for the given file

    Args:
        file_path (str): Path to the file to generate the hash for
        checksum (str): hash to apply, one of the supported by hashlib, for
            example sha1 or sha512

    Returns:
        str: hash for that file
    """

    sha = getattr(hashlib, checksum)()
    with open(file_path) as file_descriptor:
        while True:
            chunk = file_descriptor.read(65536)
            if not chunk:
                break
            sha.update(chunk)
    return sha.hexdigest()


def filter_spec(spec, paths, wildcard='*', separator='/'):
    """
    Remove keys from a spec file.
    For example, with the following path: domains/*/disks/*/metadata
    all the metadata dicts from all domains disks will be removed.

    Args:
        spec (dict): spec to remove keys from
        paths (list): list of paths to the keys that should be removed
        wildcard (str): wildcard character
        separator (str): path separator

    Returns:
        None

    Raises:
        utils.LagoUserException: If a malformed path was detected
    """

    def remove_key(path, spec):
        if len(path) == 0:
            return
        elif len(path) == 1:
            key = path.pop()
            if not isinstance(spec, collections.Mapping):
                raise LagoUserException(
                    'You have tried to remove the following key - "{key}".\n'
                    'Keys can not be removed from type {spec_type}\n'
                    'Please verify that path - "{{path}}" is valid'.format(
                        key=key, spec_type=type(spec)
                    )
                )
            if key == wildcard:
                spec.clear()
            else:
                spec.pop(key, None)
        else:
            current = path[0]
            if current == wildcard:
                if isinstance(spec, list):
                    iterator = iter(spec)
                elif isinstance(spec, collections.Mapping):
                    iterator = spec.itervalues()
                else:
                    raise LagoUserException(
                        'Glob char {char} should refer only to dict or list, '
                        'not to {spec_type}\n'
                        'Please fix path - "{{path}}"'.format(
                            char=wildcard, spec_type=type(spec)
                        )
                    )

                for i in iterator:
                    remove_key(path[1:], i)
            else:
                try:
                    remove_key(path[1:], spec[current])
                except KeyError:
                    raise LagoUserException(
                        'Malformed path "{{path}}", key "{key}" '
                        'does not exist'.format(key=current)
                    )
                except TypeError:
                    raise LagoUserException(
                        'Malformed path "{{path}}", can not get '
                        'by key from type {spec_type}'.format(
                            spec_type=type(spec)
                        )
                    )

    for path in paths:
        try:
            remove_key(path.split(separator), spec)
        except LagoUserException as e:
            e.message = e.message.format(path=path)
            raise


def ver_cmp(ver1, ver2):
    """
    Compare lago versions

    Args:
        ver1(str): version string
        ver2(str): version string

    Returns:
        Return negative if ver1<ver2, zero if ver1==ver2, positive if
        ver1>ver2.
    """

    return cmp(
        pkg_resources.parse_version(ver1), pkg_resources.parse_version(ver2)
    )


class LagoException(Exception):
    pass


class LagoInitException(LagoException):
    pass


class LagoUserException(LagoException):
    pass



#
# Copyright 2014-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
import copy
import functools
import glob
import json
import logging
import os
import shutil
import subprocess
from textwrap import dedent
import time
import urlparse
import urllib
import uuid
import warnings
import pkg_resources
from os.path import join
from plugins.output import YAMLOutFormatPlugin

import xmltodict

import paths
import subnet_lease
import utils
from utils import LagoInitException
import virt
import log_utils
import build
import sdk_utils

LOGGER = logging.getLogger(__name__)
LogTask = functools.partial(log_utils.LogTask, logger=LOGGER)
log_task = functools.partial(log_utils.log_task, logger=LOGGER)


def _create_ip(subnet, index):
    """
    Given a subnet or an ip and an index returns the ip with that lower index
    from the subnet (255.255.255.0 mask only subnets)

    Args:
        subnet (str): Strign containing the three first elements of the decimal
            representation of a subnet (X.Y.Z) or a full ip (X.Y.Z.A)
        index (int or str): Last element of a decimal ip representation, for
            example, 123 for the ip 1.2.3.123

    Returns:
        str: The dotted decimal representation of the ip
    """
    return '.'.join(subnet.split('.')[:3] + [str(index)])


def _ip_in_subnet(subnet, ip):
    """
    Checks if an ip is included in a subnet.

    Note:
        only 255.255.255.0 masks allowed

    Args:
        subnet (str): Strign containing the three first elements of the decimal
            representation of a subnet (X.Y.Z) or a full ip (X.Y.Z.A)
        ip (str or int): Decimal ip representation

    Returns:
        bool: ``True`` if ip is in subnet, ``False`` otherwise
    """
    return (
        _create_ip(subnet, 1) == _create_ip(ip, 1)
        or '0.0.0.1' == _create_ip(ip, 1)
    )


class Prefix(object):
    """
    A prefix is a directory that will contain all the data needed to setup the
    environment.

    Attributes:
        _paths (lago.path.Paths): Paths handler class
        _virt_env (lago.virt.VirtEnv): Lazily loaded virtual env handler
        _metadata (dict): Lazily loaded metadata
    """
    VIRT_ENV_CLASS = virt.VirtEnv

    def __init__(self, prefix):
        """
        Args:
            prefix (str): Path of the prefix
        """
        # self._prefix should be dropped in lago ver 0.44
        self._prefix = prefix
        self._paths = paths.Paths(prefix)
        self._virt_env = None
        self._metadata = None
        self._subnet_store = subnet_lease.SubnetStore()

    @property
    @sdk_utils.expose
    def metadata(self):
        """
        Retrieve the metadata info for this prefix

        Returns:
            dict: metadata info
        """
        if self._metadata is None:
            try:
                with open(self.paths.metadata()) as metadata_fd:
                    self._metadata = json.load(metadata_fd)
            except IOError:
                self._metadata = {}
        return self._metadata

    def _save_metadata(self):
        """
        Write this prefix metadata to disk

        Returns:
            None
        """
        with open(self.paths.metadata(), 'w') as metadata_fd:
            utils.json_dump(self.metadata, metadata_fd)

    def save(self):
        """
        Save this prefix to persistent storage

        Returns:
            None
        """
        if not os.path.exists(self.paths.virt()):
            os.makedirs(self.paths.virt())

        self._save_metadata()
        self.virt_env.save()

    def _create_ssh_keys(self):
        """
        Generate a pair of ssh keys for this prefix

        Returns:
            None

        Raises:
            RuntimeError: if it fails to create the keys
        """
        ret, _, _ = utils.run_command(
            [
                'ssh-keygen',
                '-t',
                'rsa',
                '-N',
                '',
                '-f',
                self.paths.ssh_id_rsa(),
            ]
        )
        if ret != 0:
            raise RuntimeError(
                'Failed to crate ssh keys at %s',
                self.paths.ssh_id_rsa(),
            )

    @log_task('Initialize prefix')
    def initialize(self):
        """
        Initialize this prefix, this includes creating the destination path,
        and creating the uuid for the prefix, for any other actions see
        :func:`Prefix.virt_conf`

        Will safely roll back if any of those steps fail

        Returns:
            None

        Raises:
            RuntimeError: If it fails to create the prefix dir
        """
        prefix = self.paths.prefix_path()
        os.environ['LAGO_PREFIX_PATH'] = prefix
        os.environ['LAGO_WORKDIR_PATH'] = os.path.dirname(prefix)

        with utils.RollbackContext() as rollback:
            with LogTask('Create prefix dirs'):
                try:
                    os.mkdir(prefix)
                except OSError as error:
                    raise RuntimeError(
                        'Could not create prefix at %s:\n%s' % (prefix, error)
                    )
            rollback.prependDefer(shutil.rmtree, prefix)

            with open(self.paths.uuid(), 'w') as f, \
                    LogTask('Generate prefix uuid'):
                f.write(uuid.uuid1().hex)

            with LogTask('Create ssh keys'):
                self._create_ssh_keys()

            with LogTask('Tag prefix as initialized'):
                with open(self.paths.prefix_lagofile(), 'w') as fd:
                    fd.write('')

            rollback.clear()

    @log_task('Cleanup prefix')
    def cleanup(self):
        """
        Stops any running entities in the prefix and uninitializes it, usually
        you want to do this if you are going to remove the prefix afterwards

        Returns:
            None
        """
        with LogTask('Stop prefix'):
            self.stop()
        with LogTask("Tag prefix as uninitialized"):
            os.unlink(self.paths.prefix_lagofile())

    @staticmethod
    def _init_net_specs(conf):
        """
        Given a configuration specification, initializes all the net
        definitions in it so they can be used comfortably

        Args:
            conf (dict): Configuration specification

        Returns:
            dict: the adapted new conf
        """
        for net_name, net_spec in conf.get('nets', {}).items():
            net_spec['name'] = net_name
            net_spec['mapping'] = {}
            net_spec.setdefault('type', 'nat')

        return conf

    def _allocate_subnets(self, conf):
        """
        Allocate all the subnets needed by the given configuration spec

        Args:
            conf (dict): Configuration spec where to get the nets definitions
                from

        Returns:
            tuple(list, dict): allocated subnets and modified conf
        """
        allocated_subnets = []
        try:
            for net_spec in conf.get('nets', {}).itervalues():
                if net_spec['type'] != 'nat':
                    continue

                gateway = net_spec.get('gw')
                if gateway:
                    allocated_subnet = self._subnet_store.acquire(
                        self.paths.uuid(), gateway
                    )
                else:
                    allocated_subnet = self._subnet_store.acquire(
                        self.paths.uuid()
                    )
                    net_spec['gw'] = str(allocated_subnet.iter_hosts().next())

                allocated_subnets.append(allocated_subnet)
        except:
            self._subnet_store.release(allocated_subnets)
            raise
        return allocated_subnets, conf

    def _add_nic_to_mapping(self, net, dom, nic):
        """
        Populates the given net spec mapping entry with the nics of the given
        domain, by the following rules:

            * If ``net`` is management, 'domain_name': nic_ip
            * For each interface: 'domain_name-eth#': nic_ip, where # is the
            index of the nic in the *domain* definition.
            * For each interface: 'domain_name-net_name-#': nic_ip,
            where # is a running number of interfaces from that network.
            * For each interface: 'domain_name-net_name', which has an
            identical IP to 'domain_name-net_name-0'


        Args:
            net (dict): Network spec to populate
            dom (dict): libvirt domain specification
            nic (str): Name of the interface to add to the net mapping from the
                domain

        Returns:
            None
        """
        dom_name = dom['name']
        idx = dom['nics'].index(nic)
        name = '{0}-eth{1}'.format(dom_name, idx)
        net['mapping'][name] = nic['ip']
        if dom['nics'][idx]['net'] == dom['mgmt_net']:
            net['mapping'][dom_name] = nic['ip']

        name_by_net = '{0}-{1}'.format(dom_name, nic['net'])
        named_nets = sorted(
            [
                net_name for net_name in net['mapping'].keys()
                if net_name.startswith(name_by_net) and net_name != name_by_net
            ]
        )

        if len(named_nets) == 0:
            named_idx = 0
            net['mapping'][name_by_net] = nic['ip']
        else:
            named_idx = len(named_nets)
        named_net = '{0}-{1}'.format(name_by_net, named_idx)

        net['mapping'][named_net] = nic['ip']

    def _select_mgmt_networks(self, conf):
        """
        Select management networks. If no management network is found, it will
        mark the first network found by sorted the network lists. Also adding
        default DNS domain, if none is set.

        Args:
            conf(spec): spec

        """

        nets = conf['nets']
        mgmts = sorted(
            [
                name for name, net in nets.iteritems()
                if net.get('management') is True
            ]
        )

        if len(mgmts) == 0:
            mgmt_name = sorted((nets.keys()))[0]
            LOGGER.debug(
                'No management network configured, selecting network %s',
                mgmt_name
            )
            nets[mgmt_name]['management'] = True
            mgmts.append(mgmt_name)

        for mgmt_name in mgmts:
            if nets[mgmt_name].get('dns_domain_name', None) is None:
                nets[mgmt_name]['dns_domain_name'] = 'lago.local'

        return mgmts

    def _add_dns_records(self, conf, mgmts):
        """
        Add DNS records dict('dns_records') to ``conf`` for each
        management network. Add DNS forwarder IP('dns_forward') for each none
        management network.


        Args:
            conf(spec): spec
            mgmts(list): management networks names

        Returns:
            None
        """

        nets = conf['nets']
        dns_mgmt = mgmts[-1]
        LOGGER.debug('Using network %s as main DNS server', dns_mgmt)
        forward = conf['nets'][dns_mgmt].get('gw')
        dns_records = {}
        for net_name, net_spec in nets.iteritems():
            dns_records.update(net_spec['mapping'].copy())
            if net_name not in mgmts:
                net_spec['dns_forward'] = forward

        for mgmt in mgmts:
            if nets[mgmt].get('dns_records'):
                nets[mgmt]['dns_records'].update(dns_records)
            else:
                nets[mgmt]['dns_records'] = dns_records

    def _register_preallocated_ips(self, conf):
        """
        Parse all the domains in the given conf and preallocate all their ips
        into the networks mappings, raising exception on duplicated ips or ips
        out of the allowed ranges

        See Also:
            :mod:`lago.subnet_lease`

        Args:
            conf (dict): Configuration spec to parse

        Returns:
            None

        Raises:
            RuntimeError: if there are any duplicated ips or any ip out of the
                allowed range
        """
        for dom_name, dom_spec in conf.get('domains', {}).items():
            for idx, nic in enumerate(dom_spec.get('nics', [])):
                if 'ip' not in nic:
                    continue

                net = conf['nets'][nic['net']]
                if self._subnet_store.is_leasable_subnet(net['gw']):
                    nic['ip'] = _create_ip(
                        net['gw'], int(nic['ip'].split('.')[-1])
                    )

                dom_name = dom_spec['name']
                if not _ip_in_subnet(net['gw'], nic['ip']):
                    raise RuntimeError(
                        "%s:nic%d's IP [%s] is outside the subnet [%s]" % (
                            dom_name,
                            dom_spec['nics'].index(nic),
                            nic['ip'],
                            net['gw'],
                        ),
                    )

                if nic['ip'] in net['mapping'].values():
                    conflict_list = [
                        name for name, ip in net['mapping'].items()
                        if ip == net['ip']
                    ]
                    raise RuntimeError(
                        'IP %s was to several domains: %s %s' % (
                            nic['ip'],
                            dom_name,
                            ' '.join(conflict_list),
                        ),
                    )

                self._add_nic_to_mapping(net, dom_spec, nic)

    def _get_net(self, conf, dom_name, nic):
        try:
            net = conf['nets'][nic['net']]
        except KeyError:
            raise LagoInitException(
                dedent(
                    """
                    Unrecognized network in {0}: {1},
                    available: {2}
                    """.format(
                        dom_name,
                        nic['net'],
                        ','.join(conf.get('nets', {}).keys()),
                    )
                )
            )

        return net

    def _allocate_ips_to_nics(self, conf):
        """
        For all the nics of all the domains in the conf that have dynamic ip,
        allocate one and addit to the network mapping

        Args:
            conf (dict): Configuration spec to extract the domains from

        Returns:
            None
        """
        for dom_name, dom_spec in conf.get('domains', {}).items():
            for idx, nic in enumerate(dom_spec.get('nics', [])):
                if 'ip' in nic:
                    continue
                net = self._get_net(conf, dom_name, nic)
                if net['type'] != 'nat':
                    continue

                allocated = net['mapping'].values()
                vacant = _create_ip(
                    net['gw'],
                    set(range(2, 255)).difference(
                        set([int(ip.split('.')[-1]) for ip in allocated])
                    ).pop()
                )
                nic['ip'] = vacant
                self._add_nic_to_mapping(net, dom_spec, nic)

    def _set_mtu_to_nics(self, conf):
        """
        For all the nics of all the domains in the conf that have MTU set,
        save the MTU on the NIC definition.

        Args:
            conf (dict): Configuration spec to extract the domains from

        Returns:
            None
        """
        for dom_name, dom_spec in conf.get('domains', {}).items():
            for idx, nic in enumerate(dom_spec.get('nics', [])):
                net = self._get_net(conf, dom_name, nic)
                mtu = net.get('mtu', 1500)
                if mtu != 1500:
                    nic['mtu'] = mtu

    def _config_net_topology(self, conf):
        """
        Initialize and populate all the network related elements, like
        reserving ips and populating network specs of the given confiiguration
        spec

        Args:
            conf (dict): Configuration spec to initalize

        Returns:
            None
        """
        conf = self._init_net_specs(conf)
        mgmts = self._select_mgmt_networks(conf)
        self._validate_netconfig(conf)
        allocated_subnets, conf = self._allocate_subnets(conf)
        try:
            self._add_mgmt_to_domains(conf, mgmts)
            self._register_preallocated_ips(conf)
            self._allocate_ips_to_nics(conf)
            self._set_mtu_to_nics(conf)
            self._add_dns_records(conf, mgmts)
        except:
            self._subnet_store.release(allocated_subnets)
            raise
        return conf

    def _add_mgmt_to_domains(self, conf, mgmts):
        """
        Add management network key('mgmt_net') to each domain. Note this
        assumes ``conf`` was validated.

        Args:
            conf(dict): spec
            mgmts(list): list of management networks names

        """

        for dom_name, dom_spec in conf['domains'].iteritems():
            domain_mgmt = [
                nic['net'] for nic in dom_spec['nics'] if nic['net'] in mgmts
            ].pop()

            dom_spec['mgmt_net'] = domain_mgmt

    def _validate_netconfig(self, conf):
        """
        Validate network configuration

        Args:
            conf(dict): spec

        Returns:
            None


        Raises:
            :exc:`~lago.utils.LagoInitException`: If a VM has more than
            one management network configured, or a network which is not
            management has DNS attributes, or a VM is configured with a
            none-existence NIC, or a VM has no management network.
        """

        nets = conf.get('nets', {})
        if len(nets) == 0:
            # TO-DO: add default networking if no network is configured
            raise LagoInitException('No networks configured.')

        no_mgmt_dns = [
            name for name, net in nets.iteritems()
            if net.get('management', None) is None and
            (net.get('main_dns') or net.get('dns_domain_name'))
        ]
        if len(no_mgmt_dns) > 0 and len(nets.keys()) > 1:
            raise LagoInitException(
                (
                    'Networks: {0}, misconfigured, they '
                    'are not marked as management, but have '
                    'DNS attributes. DNS is supported '
                    'only in management networks.'
                ).format(','.join(no_mgmt_dns))
            )

        for dom_name, dom_spec in conf['domains'].items():
            mgmts = []
            for nic in dom_spec['nics']:
                net = self._get_net(conf, dom_name, nic)
                if net.get('management', False) is True:
                    mgmts.append(nic['net'])
            if len(mgmts) == 0:
                raise LagoInitException(
                    (
                        'VM {0} has no management network, '
                        'please connect it to '
                        'one.'
                    ).format(dom_name)
                )

            if len(mgmts) > 1:
                raise LagoInitException(
                    (
                        'VM {0} has more than one management '
                        'network: {1}. It should have exactly '
                        'one.'
                    ).format(dom_name, ','.join(mgmts))
                )

    def _create_disk(
        self,
        name,
        spec,
        template_repo=None,
        template_store=None,
    ):
        """
        Creates a disc with the given name from the given repo or store

        Args:
            name (str): Name of the domain to create the disk for
            spec (dict): Specification of the disk to create
            template_repo (TemplateRepository or None): template repo instance
                to use
            template_store (TemplateStore or None): template store instance to
                use

        Returns:
            Tuple(str, dict): Path to the disk and disk metadata

        Raises:
            RuntimeError: If the type of the disk is not supported or failed to
                create the disk
        """
        LOGGER.debug("Spec: %s" % spec)
        with LogTask("Create disk %s" % spec['name']):
            disk_metadata = {}

            if spec['type'] == 'template':
                disk_path, disk_metadata = self._handle_template(
                    host_name=name,
                    template_spec=spec,
                    template_repo=template_repo,
                    template_store=template_store,
                )
            elif spec['type'] == 'empty':
                disk_path, disk_metadata = self._handle_empty_disk(
                    host_name=name,
                    disk_spec=spec,
                )
            elif spec['type'] == 'file':
                disk_path, disk_metadata = self._handle_file_disk(
                    disk_spec=spec,
                )
            else:
                raise RuntimeError('Unknown drive spec %s' % str(spec))

            return disk_path, disk_metadata

    def _handle_file_disk(self, disk_spec):
        url = os.path.expandvars(disk_spec.get('url', ''))
        disk_path = os.path.expandvars(disk_spec.get('path', ''))
        disk_metadata = disk_spec.get('metadata', {})
        if not url and not disk_path:
            raise RuntimeError(
                'Partial drive spec, no url nor path provided for disk of '
                'type file:\n%s' % str(disk_spec)
            )

        if url:
            disk_spec['path'] = self._retrieve_disk_url(url, disk_path)
        else:
            # Create a copy of the file or use the original
            if disk_spec.get('make_a_copy'):
                LOGGER.info("Making a copy")
                dest_path = self._generate_disk_path(
                    os.path.basename(disk_spec['path'])
                )
                # Use cp in order to keep the file sparse
                utils.cp(disk_spec['path'], dest_path)
                disk_spec['path'] = dest_path
            else:
                disk_spec['path'] = disk_path

        # If we're using raw file, return its path
        disk_path = disk_spec['path']
        return disk_path, disk_metadata

    def _retrieve_disk_url(self, disk_url, disk_dst_path=None):
        disk_in_prefix = self.fetch_url(disk_url)
        if disk_dst_path is None:
            return disk_in_prefix
        else:
            shutil.move(disk_in_prefix, disk_dst_path)
            return disk_dst_path

    @staticmethod
    def _generate_disk_name(host_name, disk_name, disk_format):
        return '%s_%s.%s' % (
            host_name,
            disk_name,
            disk_format,
        )

    def _generate_disk_path(self, disk_name):
        return os.path.expandvars(self.paths.images(disk_name))

    def _handle_empty_disk(self, host_name, disk_spec):
        disk_metadata = disk_spec.get('metadata', {})
        disk_filename = self._generate_disk_name(
            host_name=host_name,
            disk_name=disk_spec['name'],
            disk_format=disk_spec['format'],
        )
        disk_path = self._generate_disk_path(disk_name=disk_filename)

        qemu_cmd = ['qemu-img', 'create']
        if disk_spec['format'] == 'qcow2':
            qemu_cmd += [
                '-f', disk_spec['format'], '-o', 'preallocation=metadata'
            ]
        else:
            qemu_cmd += ['-f', disk_spec['format']]

        qemu_cmd += [disk_path, disk_spec['size']]

        if os.path.exists(disk_path):
            os.unlink(disk_path)

        with LogTask(
            'Create empty disk %s(%s)' % (host_name, disk_spec['name'])
        ):
            self._run_qemu(qemu_cmd, disk_path)

        disk_rel_path = os.path.join(
            '$LAGO_PREFIX_PATH',
            os.path.basename(self.paths.images()),
            os.path.basename(disk_path),
        )
        return disk_rel_path, disk_metadata

    @staticmethod
    def _run_qemu(qemu_cmd, disk_path):
        ret = utils.run_command(qemu_cmd)
        if ret.code != 0:
            raise RuntimeError(
                'Failed to create image, qemu-img returned %d:\n'
                'out:%s\nerr:%s' % ret,
            )
        # To avoid losing access to the file
        os.chmod(disk_path, 0666)
        return ret

    def _handle_template(
        self,
        host_name,
        template_spec,
        template_store=None,
        template_repo=None
    ):
        template_type = template_spec.get('template_type', 'lago')
        disk_filename = self._generate_disk_name(
            host_name=host_name,
            disk_name=template_spec['name'],
            disk_format=template_spec['format'],
        )
        disk_path = self._generate_disk_path(disk_name=disk_filename)
        if template_type == 'lago':
            qemu_cmd, disk_metadata, _ = self._handle_lago_template(
                disk_path=disk_path,
                template_spec=template_spec,
                template_store=template_store,
                template_repo=template_repo,
            )
        elif template_type == 'qcow2':
            qemu_cmd, disk_metadata = self._handle_qcow_template(
                disk_path=disk_path,
                template_spec=template_spec,
                template_store=template_store,
                template_repo=template_repo
            )
        else:
            raise RuntimeError(
                'Unsupporte template spec %s' % str(template_spec)
            )

        if os.path.exists(disk_path):
            os.unlink(disk_path)

        with LogTask(
            'Create disk %s(%s)' % (host_name, template_spec['name'])
        ):
            self._run_qemu(qemu_cmd, disk_path)

        # Update the path as relative so it can be relocated
        disk_rel_path = os.path.join(
            '$LAGO_PREFIX_PATH',
            os.path.basename(self.paths.images()),
            os.path.basename(disk_path),
        )
        return disk_rel_path, disk_metadata

    def _handle_qcow_template(
        self,
        disk_path,
        template_spec,
        template_store=None,
        template_repo=None
    ):
        base_path = template_spec.get('path', '')
        if not base_path:
            raise RuntimeError('Partial drive spec %s' % str(template_spec))

        self.resolve_parent(base_path, template_store, template_repo)

        qemu_cmd = [
            'qemu-img', 'create', '-f', 'qcow2', '-b', base_path, disk_path
        ]
        disk_metadata = template_spec.get('metadata', {})
        return qemu_cmd, disk_metadata

    def resolve_parent(self, disk_path, template_store, template_repo):
        """
        Given a virtual disk, checks if it has a backing file, if so check
        if the backing file is in the store, if not download it
        from the provided template_repo.

        After verifying that the backing-file is in the store,
        create a symlink to that file and locate it near the layered image.

        Args:
            disk_path (str): path to the layered disk
            template_repo (TemplateRepository or None): template repo instance
                to use
            template_store (TemplateStore or None): template store instance to
                use
        """
        qemu_info = utils.get_qemu_info(disk_path)
        parent = qemu_info.get('backing-filename')
        if not parent:
            return

        if os.path.isfile(parent):
            if os.path.samefile(
                os.path.realpath(parent),
                os.path.realpath(os.path.expandvars(disk_path))
            ):
                raise LagoInitException(
                    dedent(
                        """
                        Disk {} and its backing file are the same file.
                        """.format(disk_path)
                    )
                )
            # The parent exist and we have the correct pointer
            return

        LOGGER.info('Resolving Parent')
        try:
            name, version = os.path.basename(parent).split(':', 1)
        except ValueError:
            raise LagoInitException(
                dedent(
                    """
                    Backing file resolution of disk {} failed.
                    Backing file {} is not a Lago image.
                    """.format(disk_path, parent)
                )
            )

        _, _, base = self._handle_lago_template(
            '', {'template_name': name,
                 'template_version': version}, template_store, template_repo
        )

        # The child has the right pointer to his parent
        base = os.path.expandvars(base)
        if base == parent:
            return

        # The child doesn't have the right pointer to his
        # parent, We will fix it with a symlink
        link_name = os.path.join(
            os.path.dirname(disk_path), '{}:{}'.format(name, version)
        )
        link_name = os.path.expandvars(link_name)

        self._create_link_to_parent(base, link_name)

    def _create_link_to_parent(self, base, link_name):
        if not os.path.islink(link_name):
            os.symlink(base, link_name)

    def _handle_lago_template(
        self, disk_path, template_spec, template_store, template_repo
    ):
        disk_metadata = template_spec.get('metadata', {})
        if template_store is None or template_repo is None:
            raise RuntimeError('No templates directory provided')

        template = template_repo.get_by_name(template_spec['template_name'])
        template_version = template.get_version(
            template_spec.get('template_version', None)
        )
        if template_version not in template_store:
            LOGGER.info(
                log_utils.log_always("Template %s not in cache, downloading") %
                template_version.name,
            )
            template_store.download(template_version)

        disk_metadata.update(
            template_store.get_stored_metadata(
                template_version,
            ),
        )
        base = template_store.get_path(template_version)
        qemu_cmd = [
            'qemu-img', 'create', '-f', 'qcow2', '-o', 'lazy_refcounts=on',
            '-b', base, disk_path
        ]
        return qemu_cmd, disk_metadata, base

    def _ova_to_spec(self, filename):
        """
        Retrieve the given ova and makes a template of it.
        Creates a disk from network provided ova.
        Calculates the needed memory from the ovf.
        The disk will be cached in the template repo

        Args:
            filename(str): the url to retrive the data from

        TODO:
            * Add hash checking against the server
              for faster download and latest version
            * Add config script running on host - other place
            * Add cloud init support - by using cdroms in other place
            * Handle cpu in some way - some other place need to pick it up
            * Handle the memory units properly - we just assume MegaBytes

        Returns:
            list of dict: list with the disk specification
            int: VM memory, None if none defined
            int: Number of virtual cpus, None if none defined

        Raises:
            RuntimeError: If the ova format is not supported
            TypeError: If the memory units in the ova are noot supported
                (currently only 'MegaBytes')
        """
        # extract if needed
        ova_extracted_dir = os.path.splitext(filename)[0]

        if not os.path.exists(ova_extracted_dir):
            os.makedirs(ova_extracted_dir)
            subprocess.check_output(
                ["tar", "-xvf", filename, "-C", ova_extracted_dir],
                stderr=subprocess.STDOUT
            )

        # lets find the ovf file
        # we expect only one to be
        ovf = glob.glob(ova_extracted_dir + "/master/vms/*/*.ovf")
        if len(ovf) != 1:
            raise RuntimeError("We support only one vm in ova")

        image_file = None
        memory = None
        vcpus = None
        # we found our ovf
        # lets extract the resources
        with open(ovf[0]) as fd:
            # lets extract the items
            obj = xmltodict.parse(fd.read())
            hardware_items = [
                section
                for section in obj["ovf:Envelope"]["Content"]["Section"]
                if section["@xsi:type"] == "ovf:VirtualHardwareSection_Type"
            ]

            if len(hardware_items) != 1:
                raise RuntimeError("We support only one machine desc in ova")
            hardware_items = hardware_items[0]

            for item in hardware_items["Item"]:
                # lets test resource types
                CPU_RESOURCE = 3
                MEMORY_RESOURCE = 4
                DISK_RESOURCE = 17

                resource_type = int(item["rasd:ResourceType"])
                if resource_type == CPU_RESOURCE:
                    vcpus = int(item["rasd:cpu_per_socket"]) * \
                        int(item["rasd:num_of_sockets"])

                elif resource_type == MEMORY_RESOURCE:
                    memory = int(item["rasd:VirtualQuantity"])
                    if item["rasd:AllocationUnits"] != "MegaBytes":
                        raise TypeError(
                            "Fix me : we need to suport other units too"
                        )

                elif resource_type == DISK_RESOURCE:
                    image_file = item["rasd:HostResource"]

        if image_file is not None:
            disk_meta = {"root-partition": "/dev/sda1"}
            disk_spec = [
                {
                    "type": "template",
                    "template_type": "qcow2",
                    "format": "qcow2",
                    "dev": "vda",
                    "name": os.path.basename(image_file),
                    "path": ova_extracted_dir + "/images/" + image_file,
                    "metadata": disk_meta
                }
            ]

        return disk_spec, memory, vcpus

    def _use_prototype(self, spec, prototypes):
        """
        Populates the given spec with the values of it's declared prototype

        Args:
            spec (dict): spec to update
            prototypes (dict): Configuration spec containing the prototypes

        Returns:
            dict: updated spec
        """
        prototype = spec['based-on']
        del spec['based-on']
        for attr in prototype:
            if attr not in spec:
                spec[attr] = copy.deepcopy(prototype[attr])

        return spec

    def fetch_url(self, url):
        """
        Retrieves the given url to the prefix

        Args:
            url(str): Url to retrieve

        Returns:
            str: path to the downloaded file
        """
        url_path = urlparse.urlsplit(url).path
        dst_path = os.path.basename(url_path)
        dst_path = self.paths.prefixed(dst_path)
        with LogTask('Downloading %s' % url):
            urllib.urlretrieve(url=os.path.expandvars(url), filename=dst_path)

        return dst_path

    def virt_conf_from_stream(
        self,
        conf_fd,
        template_repo=None,
        template_store=None,
        do_bootstrap=True,
        do_build=True,
    ):
        """
        Initializes all the virt infrastructure of the prefix, creating the
        domains disks, doing any network leases and creating all the virt
        related files and dirs inside this prefix.

        Args:
            conf_fd (File): File like object to read the config from
            template_repo (TemplateRepository): template repository intance
            template_store (TemplateStore): template store instance

        Returns:
            None
        """
        virt_conf = utils.load_virt_stream(conf_fd)
        LOGGER.debug('Loaded virt config:\n%s', virt_conf)
        return self.virt_conf(
            conf=virt_conf,
            template_repo=template_repo,
            template_store=template_store,
            do_bootstrap=do_bootstrap,
            do_build=do_build
        )

    def _prepare_domains_images(self, conf, template_repo, template_store):
        if not os.path.exists(self.paths.images()):
            os.makedirs(self.paths.images())

        for dom_name, dom_spec in conf['domains'].items():
            if not dom_name:
                raise RuntimeError(
                    'An invalid (empty) domain name was found in the '
                    'configuration file. Cannot continue. A name must be '
                    'specified for the domain'
                )

            dom_spec['name'] = dom_name
            conf['domains'][dom_name] = self._prepare_domain_image(
                domain_spec=dom_spec,
                prototypes=conf.get('prototypes', {}),
                template_repo=template_repo,
                template_store=template_store,
            )

        return conf

    def _prepare_domain_image(
        self, domain_spec, prototypes, template_repo, template_store
    ):
        if 'based-on' in domain_spec:
            domain_spec = self._use_prototype(
                spec=domain_spec,
                prototypes=prototypes,
            )

        if domain_spec.get('type', '') == 'ova':
            domain_spec = self._handle_ova_image(domain_spec=domain_spec)

        with LogTask('Create disks for VM %s' % domain_spec['name']):
            domain_spec['disks'] = self._create_disks(
                domain_name=domain_spec['name'],
                disks_specs=domain_spec.get('disks', []),
                template_repo=template_repo,
                template_store=template_store,
            )

        return domain_spec

    def _handle_ova_image(self, domain_spec):
        # we import the ova to template
        domain_spec['type'] = 'template'
        ova_file = self.fetch_url(domain_spec['url'])
        ova_disk, domain_spec["memory"], domain_spec[
            "vcpu"
        ] = self._ova_to_spec(ova_file)
        if "disks" not in domain_spec.keys():
            domain_spec["disks"] = ova_disk
        else:
            domain_spec["disks"] = ova_disk + domain_spec["disks"]

        return domain_spec

    def _create_disks(
        self, domain_name, disks_specs, template_repo, template_store
    ):
        new_disks = []
        for disk in disks_specs:
            path, metadata = self._create_disk(
                name=domain_name,
                spec=disk,
                template_repo=template_repo,
                template_store=template_store,
            )
            new_disk = copy.deepcopy(disk)
            new_disk['path'] = path
            new_disk['metadata'] = metadata
            new_disks.append(new_disk)

        return new_disks

    def virt_conf(
        self,
        conf,
        template_repo=None,
        template_store=None,
        do_bootstrap=True,
        do_build=True
    ):
        """
        Initializes all the virt infrastructure of the prefix, creating the
        domains disks, doing any network leases and creating all the virt
        related files and dirs inside this prefix.

        Args:
            conf (dict): Configuration spec
            template_repo (TemplateRepository): template repository intance
            template_store (TemplateStore): template store instance
            do_bootstrap(bool): If true run virt-sysprep on the images
            do_build(bool): If true run build commands on the images,
                see lago.build.py for more info.

        Returns:
            None
        """
        os.environ['LAGO_PREFIX_PATH'] = self.paths.prefix_path()
        with utils.RollbackContext() as rollback:
            rollback.prependDefer(
                shutil.rmtree, self.paths.prefix_path(), ignore_errors=True
            )
            self._metadata = {
                'lago_version': pkg_resources.get_distribution("lago").version,
            }

            conf = self._prepare_domains_images(
                conf=conf,
                template_repo=template_repo,
                template_store=template_store,
            )
            conf = self._config_net_topology(conf)

            conf['domains'] = self._copy_deploy_scripts_for_hosts(
                domains=conf['domains']
            )
            self._virt_env = self.VIRT_ENV_CLASS(
                prefix=self,
                vm_specs=conf['domains'],
                net_specs=conf['nets'],
            )

            if do_bootstrap:
                self.virt_env.bootstrap()

            if do_build:
                self.build(conf['domains'])

            self.save()
            rollback.clear()

    def build(self, conf):
        builders = []
        for vm_name, spec in conf.viewitems():
            disks = spec.get('disks')
            if disks:
                for disk in disks:
                    build_spec = disk.get('build')
                    if build_spec:
                        builders.append(
                            build.Build.get_instance_from_build_spec(
                                name=vm_name,
                                disk_path=disk['path'],
                                build_spec=build_spec,
                                paths=self.paths
                            )
                        )

        utils.invoke_in_parallel(build.Build.build, builders)

    @sdk_utils.expose
    def export_vms(
        self,
        vms_names=None,
        standalone=False,
        export_dir='.',
        compress=False,
        init_file_name='LagoInitFile',
        out_format=YAMLOutFormatPlugin(),
        collect_only=False,
        with_threads=True,
    ):
        """
        Export vm images disks and init file.
        The exported images and init file can be used to recreate
        the environment.

        Args:
            vms_names(list of str): Names of the vms to export, if None
                export all the vms in the env (default=None)
            standalone(bool): If false, export a layered image
                (default=False)
            export_dir(str): Dir to place the exported images and init file
            compress(bool): If True compress the images with xz
                  (default=False)
            init_file_name(str): The name of the exported init file
                (default='LagoInitfile')
            out_format(:class:`lago.plugins.output.OutFormatPlugin`):
                The type of the exported init file (the default is yaml)
            collect_only(bool): If True, return only a mapping from vm name
                to the disks that will be exported. (default=False)
            with_threads(bool): If True, run the export in parallel
                (default=True)

        Returns
            Unless collect_only == True, a mapping between vms' disks.

        """
        return self.virt_env.export_vms(
            vms_names, standalone, export_dir, compress, init_file_name,
            out_format, collect_only, with_threads
        )

    @sdk_utils.expose
    def start(self, vm_names=None):
        """
        Start this prefix

        Args:
            vm_names(list of str): List of the vms to start

        Returns:
            None
        """
        self.virt_env.start(vm_names=vm_names)

    @sdk_utils.expose
    def stop(self, vm_names=None):
        """
        Stop this prefix

        Args:
            vm_names(list of str): List of the vms to stop

        Returns:
            None
        """
        self.virt_env.stop(vm_names=vm_names)

    @sdk_utils.expose
    def shutdown(self, vm_names=None, reboot=False):
        """
        Shutdown this prefix

        Args:
            vm_names(list of str): List of the vms to shutdown
            reboot(bool): If true, reboot the requested vms

        Returns:
            None
        """
        self.virt_env.shutdown(vm_names, reboot)

    def create_snapshots(self, name):
        """
        Creates one snapshot on all the domains with the given name

        Args:
            name (str): Name of the snapshots to create

        Returns:
            None
        """
        self.virt_env.create_snapshots(name)

    def revert_snapshots(self, name):
        """
        Revert all the snapshots with the given name from all the domains

        Args:
            name (str): Name of the snapshots to revert

        Returns:
            None
        """
        self.virt_env.revert_snapshots(name)

    def get_snapshots(self):
        """
        Retrieve info on all the snapshots from all the domains

        Returns:
            dict of str: list(str): dictionary with vm_name -> snapshot list
        """
        return self.virt_env.get_snapshots()

    def _create_virt_env(self):
        """
        Create a new virt env from this prefix

        Returns:
            lago.virt.VirtEnv: virt env created from this prefix
        """
        return self.VIRT_ENV_CLASS.from_prefix(self)

    @property
    def virt_env(self):
        """
        Getter for this instance's virt env, creates it if needed

        Returns:
            lago.virt.VirtEnv: virt env instance used by this prefix
        """
        if self._virt_env is None:
            self._virt_env = self._create_virt_env()
        return self._virt_env

    @property
    def paths(self):
        return self._paths

    @paths.setter
    def paths(self, val):
        self._paths = val

    def destroy(self):
        """
        Destroy this prefix, running any cleanups and removing any files
        inside it.
        """

        subnets = (
            str(net.gw()) for net in self.virt_env.get_nets().itervalues()
        )

        self._subnet_store.release(subnets)
        self.cleanup()
        shutil.rmtree(self.paths.prefix_path())

    @sdk_utils.expose
    def get_vms(self):
        """
        Retrieve info on all the vms

        Returns:
            dict of str->list(str): dictionary with vm_name -> vm list
        """
        return self.virt_env.get_vms()

    @sdk_utils.expose
    def get_nets(self):
        """
        Retrieve info on all the nets from all the domains

        Returns:
            dict of str->list(str): dictionary with net_name -> net list
        """
        return self.virt_env.get_nets()

    @sdk_utils.expose
    def get_paths(self):
        """
        Get the Paths object of this prefix.
        The Paths object contains the paths to the internal directories
        and files of this prefix.

        Returns:
            :class:`lago.paths.Paths`: The Paths object of this prefix
        """
        return self.paths

    @classmethod
    def resolve_prefix_path(cls, start_path=None):
        """
        Look for an existing prefix in the given path, in a path/.lago dir, or
        in a .lago dir under any of it's parent directories

        Args:
            start_path (str): path to start the search from, if None passed, it
                will use the current dir

        Returns:
            str: path to the found prefix

        Raises:
            RuntimeError: if no prefix was found
        """
        if not start_path or start_path == 'auto':
            start_path = os.path.curdir

        cur_path = start_path
        LOGGER.debug('Checking if %s is a prefix', os.path.abspath(cur_path))
        if cls.is_prefix(cur_path):
            return os.path.abspath(cur_path)

        # now search for a .lago directory that's a prefix on any parent dir
        cur_path = join(start_path, '.lago')
        while not cls.is_prefix(cur_path):
            LOGGER.debug('%s  is not a prefix', cur_path)
            cur_path = os.path.normpath(
                os.path.join(cur_path, '..', '..', '.lago')
            )
            LOGGER.debug('Checking %s for a prefix', cur_path)
            if os.path.realpath(join(cur_path, '..')) == '/':
                raise RuntimeError(
                    'Unable to find prefix for %s' %
                    os.path.abspath(start_path)
                )

        return os.path.abspath(cur_path)

    @classmethod
    def is_prefix(cls, path):
        """
        Check if a path is a valid prefix

        Args:
            path(str): path to be checked

        Returns:
            bool: True if the given path is a prefix
        """
        lagofile = paths.Paths(path).prefix_lagofile()
        return os.path.isfile(lagofile)

    @sdk_utils.expose
    @log_task('Collect artifacts')
    def collect_artifacts(self, output_dir, ignore_nopath):
        if os.path.exists(output_dir):
            utils.rotate_dir(output_dir)

        os.makedirs(output_dir)

        def _collect_artifacts(vm):
            with LogTask('%s' % vm.name()):
                path = os.path.join(output_dir, vm.name())
                os.makedirs(path)
                vm.collect_artifacts(path, ignore_nopath)

        utils.invoke_in_parallel(
            _collect_artifacts,
            self.virt_env.get_vms().values(),
        )

    def _get_scripts(self, host_metadata):
        """
        Temporary method to retrieve the host scripts

        TODO:
            remove once the "ovirt-scripts" option gets deprecated

        Args:
            host_metadata(dict): host metadata to retrieve the scripts for

        Returns:
            list: deploy scripts for the host, empty if none found
        """
        deploy_scripts = host_metadata.get('deploy-scripts', [])
        if deploy_scripts:
            return deploy_scripts

        ovirt_scripts = host_metadata.get('ovirt-scripts', [])
        if ovirt_scripts:
            warnings.warn(
                'Deprecated entry "ovirt-scripts" will not be supported in '
                'the future, replace with "deploy-scripts"'
            )

        return ovirt_scripts

    def _set_scripts(self, host_metadata, scripts):
        """
        Temporary method to set the host scripts

        TODO:
            remove once the "ovirt-scripts" option gets deprecated

        Args:
            host_metadata(dict): host metadata to set scripts in

        Returns:
            dict: the updated metadata
        """
        scripts_key = 'deploy-scripts'
        if 'ovirt-scritps' in host_metadata:
            scripts_key = 'ovirt-scripts'

        host_metadata[scripts_key] = scripts
        return host_metadata

    def _copy_deploy_scripts_for_hosts(self, domains):
        """
        Copy the deploy scripts for all the domains into the prefix scripts dir

        Args:
            domains(dict): spec with the domains info as when loaded from the
                initfile

        Returns:
            None
        """
        with LogTask('Copying any deploy scripts'):
            for host_name, host_spec in domains.iteritems():
                host_metadata = host_spec.get('metadata', {})
                deploy_scripts = self._get_scripts(host_metadata)
                new_scripts = self._copy_delpoy_scripts(deploy_scripts)
                self._set_scripts(
                    host_metadata=host_metadata,
                    scripts=new_scripts,
                )

        return domains

    def _copy_delpoy_scripts(self, scripts):
        """
        Copy the given deploy scripts to the scripts dir in the prefix

        Args:
            scripts(list of str): list of paths of the scripts to copy to the
                prefix

        Returns:
            list of str: list with the paths to the copied scripts, with a
                prefixed with $LAGO_PREFIX_PATH so the full path is not
                hardcoded
        """
        if not os.path.exists(self.paths.scripts()):
            os.makedirs(self.paths.scripts())

        new_scripts = []
        for script in scripts:
            script = os.path.expandvars(script)
            if not os.path.exists(script):
                raise RuntimeError('Script %s does not exist' % script)

            sanitized_name = script.replace('/', '_')
            new_script_cur_path = os.path.expandvars(
                self.paths.scripts(sanitized_name)
            )
            shutil.copy(script, new_script_cur_path)

            new_script_init_path = os.path.join(
                '$LAGO_PREFIX_PATH',
                os.path.basename(self.paths.scripts()),
                sanitized_name,
            )
            new_scripts.append(new_script_init_path)

        return new_scripts

    def _deploy_host(self, host):
        with LogTask('Deploy VM %s' % host.name()):
            deploy_scripts = self._get_scripts(host.metadata)
            if not deploy_scripts:
                return

            with LogTask('Wait for ssh connectivity'):
                if not host.ssh_reachable(tries=1, propagate_fail=False):
                    time.sleep(10)
                    host.wait_for_ssh()

            for script in deploy_scripts:
                script = os.path.expanduser(os.path.expandvars(script))
                with LogTask('Run script %s' % os.path.basename(script)):
                    ret, out, err = host.ssh_script(script, show_output=False)

                if ret != 0:
                    LOGGER.debug('STDOUT:\n%s' % out)
                    LOGGER.error('STDERR\n%s' % err)
                    raise RuntimeError(
                        '%s failed with status %d on %s' % (
                            script,
                            ret,
                            host.name(),
                        ),
                    )

    @sdk_utils.expose
    @log_task('Deploy environment')
    def deploy(self):
        utils.invoke_in_parallel(
            self._deploy_host,
            self.virt_env.get_vms().values()
        )



# encoding: utf-8
"""
This module contains any disk template related classes and functions, including
the repository store manager classes and template providers, some useful
definitions:

    * Template repositories:
        Repository where to fetch templates from, as an http server

    * Template store:
        Local store to cache templates

    * Template:
        Unititialized disk image to use as base for other disk images

    * Template version:
        Specific version of a template, to allow getting updates without
        having to change the template name everywhere

"""
import errno
import functools
import json
import logging
import os
import posixpath
import shutil
import urllib
import sys

import lockfile

import utils
from . import log_utils
from .config import config

LOGGER = logging.getLogger(__name__)


class FileSystemTemplateProvider:
    """
    Handles file type templates, that is, getting a disk template from the
    host's filesystem
    """

    def __init__(self, root):
        """
        Args:
            root (str): Path to the template, any vars and user globs wil be
                expanded
        """
        self._root = os.path.expanduser(os.path.expandvars(root))

    def _prefixed(self, *path):
        """
        Join all the given paths prefixed with this provider's base root path

        Args:
            *path (str): sections of the path to join, passed as positional
                arguments

        Returns:
            str: Joined paths prepended with the provider root path
        """
        return os.path.join(self._root, *path)

    def download_image(self, handle, dest):
        """
        Copies over the handl to the destination

        Args:
            handle (str): path to copy over
            dest (str): path to copy to

        Returns:
            None
        """
        shutil.copyfile(self._prefixed(handle), dest)

    def get_hash(self, handle):
        """
        Returns the associated hash for the given handle, the hash file must
        exist (``handle + '.hash'``).

        Args:
            handle (str): Path to the template to get the hash from

        Returns:
            str: Hash for the given handle
        """
        handle = os.path.expanduser(os.path.expandvars(handle))
        with open(self._prefixed('%s.hash' % handle)) as f:
            return f.read()

    def get_metadata(self, handle):
        """
        Returns the associated metadata info for the given handle, the metadata
        file must exist (``handle + '.metadata'``).

        Args:
            handle (str): Path to the template to get the metadata from

        Returns:
            dict: Metadata for the given handle
        """
        handle = os.path.expanduser(os.path.expandvars(handle))
        with open(self._prefixed('%s.metadata' % handle)) as f:
            return json.load(f)


class HttpTemplateProvider:
    """
    This provider allows the usage of http urls for templates
    """

    def __init__(self, baseurl):
        """
        Args:
           baseurl (str): Url to prepend to every handle
        """
        self.baseurl = baseurl

    def open_url(self, url, suffix='', dest=None):
        """
        Opens the given url, trying the compressed version first.
        The compressed version url is generated adding the ``.xz`` extension
        to the ``url`` and adding the given suffix **after** that ``.xz``
        extension.
        If dest passed, it will download the data to that path if able

        Args:
            url (str): relative url from the ``self.baseurl`` to retrieve
            suffix (str): optional suffix to append to the url after adding
                the compressed extension to the path
            dest (str or None): Path to save the data to

        Returns:
            urllib.addinfourl: response object to read from (lazy read), closed
                if no dest passed

        Raises:
            RuntimeError: if the url gave http error when retrieving it
        """
        if not url.endswith('.xz'):
            try:
                return self.open_url(
                    url=url + '.xz',
                    suffix=suffix,
                    dest=dest,
                )
            except RuntimeError:
                pass
        full_url = posixpath.join(self.baseurl, url) + suffix
        response = urllib.urlopen(full_url)
        if response.code >= 300:
            raise RuntimeError(
                'Failed no retrieve URL %s:\nCode: %d' %
                (full_url, response.code)
            )

        meta = response.info()
        file_size_kb = int(meta.getheaders("Content-Length")[0]) / 1024
        if file_size_kb > 0:
            sys.stdout.write(
                "Downloading %s Kilobytes from %s \n" %
                (file_size_kb, full_url)
            )

        def report(count, block_size, total_size):
            percent = (count * block_size * 100 / float(total_size))
            sys.stdout.write(
                "\r% 3.1f%%" % percent + " complete (%d " %
                (count * block_size / 1024) + "Kilobytes)"
            )
            sys.stdout.flush()

        if dest:
            response.close()
            urllib.urlretrieve(full_url, dest, report)
            sys.stdout.write("\n")
        return response

    def download_image(self, handle, dest):
        """
        Downloads the image from the http server

        Args:
            handle (str): url from the `self.baseurl` to the remote template
            dest (str): Path to store the downloaded url to, must be a file
                path

        Returns:
            None
        """
        with log_utils.LogTask('Download image %s' % handle, logger=LOGGER):
            self.open_url(url=handle, dest=dest)

        self.extract_image_xz(dest)

    @staticmethod
    def extract_image_xz(path):
        if not path.endswith('.xz'):
            os.rename(path, path + '.xz')
            path = path + '.xz'

        with log_utils.LogTask('Decompress local image', logger=LOGGER):
            ret = utils.run_command(
                ['xz', '--threads=0', '--decompress', path],
            )

        if ret:
            raise RuntimeError('Failed to decompress %s' % path)

    def get_hash(self, handle):
        """
        Get the associated hash for the given handle, the hash file must
        exist (``handle + '.hash'``).

        Args:
            handle (str): Path to the template to get the hash from

        Returns:
            str: Hash for the given handle
        """
        response = self.open_url(url=handle, suffix='.hash')
        try:
            return response.read()
        finally:
            response.close()

    def get_metadata(self, handle):
        """
        Returns the associated metadata info for the given handle, the metadata
        file must exist (``handle + '.metadata'``). If the given handle has an
        ``.xz`` extension, it will get removed when calculating the handle
        metadata path

        Args:
            handle (str): Path to the template to get the metadata from

        Returns:
            dict: Metadata for the given handle
        """
        response = self.open_url(url=handle, suffix='.metadata')
        try:
            return json.load(response)
        finally:
            response.close()


#: Registry for template providers
_PROVIDERS = {
    'file': FileSystemTemplateProvider,
    'http': HttpTemplateProvider,
}


def find_repo_by_name(name, repo_dir=None):
    """
    Searches the given repo name inside the repo_dir (will use the config value
    'template_repos' if no repo dir passed), will rise an exception if not
    found

    Args:
        name (str): Name of the repo to search
        repo_dir (str): Directory where to search the repo

    Return:
        str: path to the repo

    Raises:
        RuntimeError: if not found
    """
    if repo_dir is None:
        repo_dir = config.get('template_repos')

    ret, out, _ = utils.run_command([
        'find',
        repo_dir,
        '-name',
        '*.json',
    ], )

    repos = [
        TemplateRepository.from_url(line.strip()) for line in out.split('\n')
        if len(line.strip())
    ]

    for repo in repos:
        if repo.name == name:
            return repo
    raise RuntimeError('Could not find repo %s' % name)


class TemplateRepository:
    """
    A template repository is a single source for templates, that uses different
    providers to actually retrieve them. That means for example that the
    'ovirt' template repository, could support the 'http' and a theoretical
    'gluster' template providers.


    Attributes:
        _dom (dict): Specification of the template
        _providers (dict): Providers instances for any source in the spec
    """

    def __init__(self, dom):
        """
        You would usually use the
        :func:`TemplateRepository.from_url` method instead of
        directly using this

        Args:
            dom (dict): Specification of the template repository (not confuse
                with xml dom)
        """
        self._dom = dom
        self._providers = {
            name: self._get_provider(spec)
            for name, spec in self._dom.get('sources', {}).items()
        }

    @classmethod
    def from_url(cls, path):
        """
        Instantiate a :class:`TemplateRepository` instance from the data in a
        file or url

        Args:
            path (str): Path or url to the json file to load

        Returns:
            TemplateRepository: A new instance
        """
        if os.path.isfile(path):
            with open(path) as fd:
                data = fd.read()
        else:
            try:
                response = urllib.urlopen(path)
                if response.code >= 300:
                    raise RuntimeError('Unable to load repo from %s' % path)

                data = response.read()
                response.close()
            except IOError:
                raise RuntimeError(
                    'Unable to load repo from %s (IO error)' % path
                )

        return cls(json.loads(data))

    def _get_provider(self, spec):
        """
        Get the provider for the given template spec

        Args:
            spec (dict): Template spec

        Returns:
            HttpTemplateProvider or FileSystemTemplateProvider:
                A provider instance for that spec
        """
        provider_class = _PROVIDERS[spec['type']]
        return provider_class(**spec['args'])

    @property
    def name(self):
        """
        Getter for the template repo name

        Returns:
            str: the name of this template repo
        """
        return self._dom['name']

    def get_by_name(self, name):
        """
        Retrieve a template by it's name

        Args:
            name (str): Name of the template to retrieve

        Raises:
            KeyError: if no template is found
        """
        spec = self._dom.get('templates', {})[name]
        return Template(
            name=name,
            versions={
                ver_name: TemplateVersion(
                    name='%s:%s:%s' % (self.name, name, ver_name),
                    source=self._providers[ver_spec['source']],
                    handle=ver_spec['handle'],
                    timestamp=ver_spec['timestamp'],
                )
                for ver_name, ver_spec in spec['versions'].items()
            },
        )


class Template:
    """
    Disk image template class

    Attributes:
        name (str): Name of this template
        _versions (dict(str:TemplateVersion)): versions for this template
    """

    def __init__(self, name, versions):
        """
        Args:
            name (str): Name of the template
            versions (dict(str:TemplateVersion)): dictionary with the
                version_name: :class:`TemplateVersion` pairs for this template
        """
        self.name = name
        self._versions = versions

    def get_version(self, ver_name=None):
        """
        Get the given version for this template, or the latest

        Args:
            ver_name (str or None): Version to retieve, None for the latest

        Returns:
            TemplateVersion: The version matching the given name or the latest
                one
        """
        if ver_name is None:
            return self.get_latest_version()
        return self._versions[ver_name]

    def get_latest_version(self):
        """
        Retrieves the latest version for this template, the latest being the
        one with the newest timestamp

        Returns:
            TemplateVersion
        """
        return max(self._versions.values(), key=lambda x: x.timestamp())


class TemplateVersion:
    """
    Each template can have multiple versions, each of those is actually a
    different disk template file representation, under the same base name.
    """

    def __init__(self, name, source, handle, timestamp):
        """

        Args:
            name (str): Base name of the template
            source (HttpTemplateProvider or FileSystemTemplateProvider):
                template provider for this version
            handle (str): handle of the template version, this is the
                information that will be used passed to the repo provider to
                retrieve the template (depends on the provider)
            timestamp (int): timestamp as seconds since 1970-01-01 00:00:00
                UTC
        """
        self.name = name
        self._source = source
        self._handle = handle
        self._timestamp = timestamp
        self._hash = None
        self._metadata = None

    def timestamp(self):
        """
        Getter for the timestamp
        """
        return self._timestamp

    def get_hash(self):
        """
        Returns the associated hash for this template version

        Returns:
            str: Hash for this version
        """
        if self._hash is None:
            self._hash = self._source.get_hash(self._handle).strip()
        return self._hash

    def get_metadata(self):
        """
        Returns the associated metadata info for this template version

        Returns:
            dict: Metadata for this version
        """

        if self._metadata is None:
            self._metadata = self._source.get_metadata(self._handle)
        return self._metadata

    def download(self, destination):
        """
        Retrieves this template to the destination file

        Args:
            destination (str): file path to write this template to

        Returns:
            None
        """
        self._source.download_image(self._handle, destination)


def _locked(func):
    """
    Decorator that ensures that the decorated function has the lock of the
    repo while running, meant to decorate only bound functions for classes that
    have `lock_path` method.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        with lockfile.LockFile(self.lock_path()):
            return func(self, *args, **kwargs)


class TemplateStore:
    """
    Local cache to store templates

    The store uses various files to keep track of the templates cached, access
    and versions. An example template store looks like::

        $ tree /var/lib/lago/store/
        /var/lib/lago/store/
        ├── in_office_repo:centos6_engine:v2.tmp
        ├── in_office_repo:centos7_engine:v5.tmp
        ├── in_office_repo:fedora22_host:v2.tmp
        ├── phx_repo:centos6_engine:v2
        ├── phx_repo:centos6_engine:v2.hash
        ├── phx_repo:centos6_engine:v2.metadata
        ├── phx_repo:centos6_engine:v2.users
        ├── phx_repo:centos7_engine:v4.tmp
        ├── phx_repo:centos7_host:v4.tmp
        └── phx_repo:storage-nfs:v1.tmp

    There you can see the files:

    * \*.tmp
        Temporary file created while downloading the template from the
        repository (depends on the provider)

    * ${repo_name}:${template_name}:${template_version}
        This file is the actual disk image template

    * \*.hash
        Cached hash for the template disk image

    * \*.metadata
        Metadata for this template image in json format, usually this includes
        the `distro` and `root-password`
    """

    def __init__(self, path):
        """
        :param str path: Path to a local dir for this store, will be created if
            it does not exist
        :raises OSError: if there's a failure creating the dir
        """
        self._root = path
        try:
            os.makedirs(path)
        except OSError as e:
            if e.errno != errno.EEXIST:
                LOGGER.error('Failed to create store dir')
                raise

    def _prefixed(self, *path):
        """
        Join the given paths and prepend this stores path

        Args:
            *path (str): list of paths to join, as positional arguments

        Returns:
            str: all the paths joined and prepended with the store path
        """
        return os.path.join(self._root, *path)

    def __contains__(self, temp_ver):
        """
        Checks if a given version is in this store

        Args:
            temp_ver (TemplateVersion): Version to look for

        Returns:
            bool: ``True`` if the version is in this store
        """
        return os.path.exists(self._prefixed(temp_ver.name))

    def get_path(self, temp_ver):
        """
        Get the path of the given version in this store

        Args:
            temp_ver TemplateVersion: version to look for

        Returns:
            str: The path to the template version inside the store

        Raises:
            RuntimeError: if the template is not in the store
        """
        if temp_ver not in self:
            raise RuntimeError(
                'Template: {} not present'.format(temp_ver.name)
            )
        return self._prefixed(temp_ver.name)

    def download(self, temp_ver, store_metadata=True):
        """
        Retrieve the given template version

        Args:
            temp_ver (TemplateVersion): template version to retrieve
            store_metadata (bool): If set to ``False``, will not refresh the
                local metadata with the retrieved one

        Returns:
            None
        """
        dest = self._prefixed(temp_ver.name)
        temp_dest = '%s.tmp' % dest

        with lockfile.LockFile(dest):
            # Image was downloaded while we were waiting
            if os.path.exists(dest):
                return

            temp_ver.download(temp_dest)
            if store_metadata:
                with open('%s.metadata' % dest, 'w') as f:
                    utils.json_dump(temp_ver.get_metadata(), f)

            sha1 = utils.get_hash(temp_dest)
            if temp_ver.get_hash() != sha1:
                raise RuntimeError(
                    'Image %s does not match the expected hash %s' % (
                        temp_ver.name,
                        sha1.hexdigest(),
                    )
                )

            with open('%s.hash' % dest, 'w') as f:
                f.write(sha1)

            with log_utils.LogTask('Convert image', logger=LOGGER):
                result = utils.run_command(
                    [
                        'qemu-img',
                        'convert',
                        '-O',
                        'raw',
                        temp_dest,
                        dest,
                    ],
                )

                os.unlink(temp_dest)
                if result:
                    raise RuntimeError(result.err)

    def get_stored_metadata(self, temp_ver):
        """
        Retrieves the metadata for the given template version from the store

        Args:
            temp_ver (TemplateVersion): template version to retrieve the
                metadata for

        Returns:
            dict: the metadata of the given template version
        """
        with open(self._prefixed('%s.metadata' % temp_ver.name)) as f:
            return json.load(f)

    def get_stored_hash(self, temp_ver):
        """
        Retrieves the hash for the given template version from the store

        Args:
            temp_ver (TemplateVersion): template version to retrieve the hash
                for

        Returns:
            str: hash of the given template version
        """
        with open(self._prefixed('%s.hash' % temp_ver.name)) as f:
            return f.read().strip()



#
# Copyright 2016-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
"""
A workdir is the base directory where lago will store all the files it needs
and that are unique (not shared between workdirs).

It's basic structure is a directory with one soft link and multiple
directories, one per prefix. Where the link points to the default prefix to
use.

"""
import os
import logging
import shutil
from functools import partial, wraps

from . import (prefix, utils)
from .plugins import cli
from .utils import LagoUserException

LOGGER = logging.getLogger(__name__)


class WorkdirError(RuntimeError):
    """
    Base exception for workdir errors, catch this one to catch any workdir
    error
    """
    pass


class PrefixNotFound(WorkdirError):
    pass


class MalformedWorkdir(WorkdirError):
    pass


class PrefixAlreadyExists(WorkdirError):
    pass


def workdir_loaded(func):
    """
    Decorator to make sure that the workdir is loaded when calling the
    decorated function
    """

    @wraps(func)
    def decorator(workdir, *args, **kwargs):
        if not workdir.loaded:
            workdir.load()

        return func(workdir, *args, **kwargs)

    return decorator


class Workdir(object):
    """
    This class reperesents a base workdir, where you can store multiple
    prefixes

    Properties:
        path(str): Path to the workdir
        perfixes(dict of str->self.prefix_class): dict with the prefixes in
        the workdir, by name
        current(str): Name of the current prefix
        prefix_class(type): Class to use when creating prefixes
    """

    def __init__(self, path, prefix_class=prefix.Prefix):
        self.path = path
        self.prefixes = {}
        self.current = None
        self.loaded = False
        self.prefix_class = prefix_class

    def join(self, *args):
        """
        Gets a joined path prefixed with the workdir path

        Args:
            *args(str): path sections to join

        Returns:
            str: Joined path prefixed with the workdir path
        """
        return os.path.join(self.path, *args)

    def initialize(self, prefix_name='default', *args, **kwargs):
        """
        Initializes a workdir by adding a new prefix to the workdir.

        Args:
            prefix_name(str): Name of the new prefix to add
            *args: args to pass along to the prefix constructor
            *kwargs: kwargs to pass along to the prefix constructor

        Returns:
            The newly created prefix

        Raises:
            PrefixAlreadyExists: if the prefix name already exists in the
                workdir
        """
        if self.loaded:
            raise WorkdirError('Workdir %s already initialized' % self.path)

        if not os.path.exists(self.path):
            LOGGER.debug('Creating workdir %s', self.path)
            os.makedirs(self.path)

        self.prefixes[prefix_name] = self.prefix_class(
            self.join(prefix_name), *args, **kwargs
        )
        self.prefixes[prefix_name].initialize()
        if self.current is None:
            self._set_current(prefix_name)

        self.load()

        return self.prefixes[prefix_name]

    def load(self):
        """
        Loads the prefixes that are available is the workdir

        Returns:
            None

        Raises:
            MalformedWorkdir: if the wordir is malformed
        """
        if self.loaded:
            LOGGER.debug('Already loaded')
            return

        try:
            basepath, dirs, _ = os.walk(self.path).next()
        except StopIteration:
            raise MalformedWorkdir('Empty dir %s' % self.path)

        full_path = partial(os.path.join, basepath)
        found_current = False

        for dirname in dirs:
            if dirname == 'current' and os.path.islink(full_path('current')):
                self.current = os.path.basename(
                    os.readlink(full_path('current'))
                )
                found_current = True
                continue
            elif dirname == 'current':
                raise MalformedWorkdir(
                    '"%s/current" should be a soft link' % self.path
                )

            self.prefixes[dirname] = self.prefix_class(
                prefix=self.join(dirname)
            )

        if not found_current:
            raise MalformedWorkdir(
                '"%s/current" should exist and be a soft link' % self.path
            )

        self._update_current()

    def _update_current(self):
        """
        Makes sure that a current is set
        """
        if not self.current or self.current not in self.prefixes:
            if 'default' in self.prefixes:
                selected_current = 'default'
            elif self.prefixes:
                selected_current = sorted(self.prefixes.keys()).pop()
            else:
                # should never get here
                raise MalformedWorkdir(
                    'No current link and no prefixes in workdir %s' % self.path
                )

            logging.info(
                'Missing current link, setting it to %s',
                selected_current,
            )
            self._set_current(selected_current)

    def _set_current(self, new_current):
        """
        Change the current default prefix, for internal usage

        Args:
            new_current(str): Name of the new current prefix, it must already
                exist

        Returns:
            None

        Raises:
            PrefixNotFound: if the given prefix name does not exist in the
                workdir
        """
        new_cur_full_path = self.join(new_current)
        if not os.path.exists(new_cur_full_path):
            raise PrefixNotFound(
                'Prefix "%s" does not exist in workdir %s' %
                (new_current, self.path)
            )

        if os.path.lexists(self.join('current')):
            os.unlink(self.join('current'))

        os.symlink(new_current, self.join('current'))
        self.current = new_current

    @workdir_loaded
    def set_current(self, new_current):
        """
        Change the current default prefix

        Args:
            new_current(str): Name of the new current prefix, it must already
                exist

        Returns:
            None

        Raises:
            PrefixNotFound: if the given prefix name does not exist in the
                workdir
        """
        self._set_current(new_current)

    @workdir_loaded
    def add_prefix(self, name, *args, **kwargs):
        """
        Adds a new prefix to the workdir.

        Args:
            name(str): Name of the new prefix to add
            *args: args to pass along to the prefix constructor
            *kwargs: kwargs to pass along to the prefix constructor

        Returns:
            The newly created prefix

        Raises:
            PrefixAlreadyExists: if the prefix name already exists in the
                workdir
        """
        if os.path.exists(self.join(name)):
            raise PrefixAlreadyExists(
                'Prefix with name %s already exists in workdir %s' %
                (name, self.path)
            )

        self.prefixes[name] = self.prefix_class(
            self.join(name), *args, **kwargs
        )
        self.prefixes[name].initialize()
        if self.current is None:
            self.set_current(name)

        return self.prefixes[name]

    @workdir_loaded
    def get_prefix(self, name):
        """
        Retrieve a prefix, resolving the current one if needed

        Args:
            name(str): name of the prefix to retrieve, or current to get the
                current one

        Returns:
            self.prefix_class: instance of the prefix with the given name
        """
        if name == 'current':
            name = self.current

        try:
            return self.prefixes[name]
        except KeyError:
            raise KeyError(
                'Unable to find prefix "%s" in workdir %s' % (name, self.path)
            )

    @workdir_loaded
    def destroy(self, prefix_names=None):
        """
        Destroy all the given prefixes and remove any left files if no more
        prefixes are left

        Args:
            prefix_names(list of str): list of prefix names to destroy, if None
            passed (default) will destroy all of them
        """
        if prefix_names is None:
            self.destroy(prefix_names=self.prefixes.keys())
            return

        for prefix_name in prefix_names:
            if prefix_name == 'current' and self.current in prefix_names:
                continue

            elif prefix_name == 'current':
                prefix_name = self.current

            self.get_prefix(prefix_name).destroy()
            self.prefixes.pop(prefix_name)
            if self.prefixes:
                self._update_current()

        if not self.prefixes:
            shutil.rmtree(self.path)

    @classmethod
    def resolve_workdir_path(cls, start_path=os.curdir):
        """
        Look for an existing workdir in the given path, in a path/.lago dir,
        or in a .lago dir under any of it's parent directories

        Args:
            start_path (str): path to start the search from, if None passed, it
                will use the current dir

        Returns:
            str: path to the found prefix

        Raises:
            LagoUserException: if no prefix was found
        """
        if start_path == 'auto':
            start_path = os.curdir

        cur_path = start_path

        LOGGER.debug(
            'Checking if %s is a workdir',
            os.path.abspath(cur_path),
        )
        if cls.is_workdir(cur_path):
            return os.path.abspath(cur_path)

        # now search for a .lago directory that's a workdir on any parent dir
        cur_path = os.path.join(start_path, '.lago')
        while not cls.is_workdir(cur_path):
            LOGGER.debug('%s is not a workdir', cur_path)
            cur_path = os.path.normpath(
                os.path.join(cur_path, '..', '..', '.lago')
            )
            LOGGER.debug('Checking %s for a workdir', cur_path)
            if os.path.realpath(os.path.join(cur_path, '..')) == '/':
                # no workdir found - look workdirs up the current path + 1,
                # print informative message and exit.
                candidates = []
                for path in os.listdir(os.curdir):
                    if os.path.isdir(path):
                        dirs = os.listdir(path)
                        if 'current' in dirs:
                            candidates.append(
                                os.path.abspath(os.path.join(os.curdir, path))
                            )
                        elif '.lago' in dirs:
                            candidates.append(
                                os.path.abspath(
                                    os.path.join(os.curdir, path, '.lago')
                                )
                            )
                candidates = filter(Workdir.is_possible_workdir, candidates)
                for idx in range(len(candidates)):
                    if os.path.split(candidates[idx])[1] == '.lago':
                        candidates[idx] = os.path.dirname(candidates[idx])

                msg = 'Unable to find workdir in {0}'.format(
                    os.path.abspath(start_path)
                )
                if candidates:
                    msg += '\nFound possible workdirs in: {0}'.format(
                        ', '.join(candidates)
                    )
                raise LagoUserException(msg)

        return os.path.abspath(cur_path)

    @staticmethod
    def is_possible_workdir(path):
        """
        A quick method to suggest if the path is a possible workdir.
        This does not guarantee that the workdir is not malformed, only that by
        simple heuristics it might be one.
        For a full check use :func:`is_workdir`.

        Args:
            path(str): Path

        Returns:
            bool: True if ``path`` might be a work dir.
        """
        res = False
        trails = ['initialized', 'uuid']
        try:
            res = all(
                os.path.isfile(os.path.join(path, 'current', trail))
                for trail in trails
            )
        except:
            pass
        return res

    @classmethod
    def is_workdir(cls, path):
        """
        Check if the given path is a workdir

        Args:
            path(str): Path to check

        Return:
            bool: True if the given path is a workdir
        """
        try:
            cls(path=path).load()
        except MalformedWorkdir:
            return False

        return True

    def cleanup(self):
        """
        Attempt to set a new current symlink if it is broken. If no other
        prefixes exist and the workdir is empty, try to delete the entire
        workdir.

        Raises:
            :exc:`~MalformedWorkdir`: if no prefixes were found, but the
                workdir is not empty.
        """

        current = self.join('current')
        if not os.path.exists(current):
            LOGGER.debug('found broken current symlink, removing: %s', current)
            os.unlink(self.join('current'))
            self.current = None
            try:
                self._update_current()
            except PrefixNotFound:
                if not os.listdir(self.path):
                    LOGGER.debug('workdir is empty, removing %s', self.path)
                    os.rmdir(self.path)
                else:
                    raise MalformedWorkdir(
                        (
                            'Unable to find any prefixes in {0}, '
                            'but the directory looks malformed. '
                            'Try deleting it manually.'
                        ).format(self.path)
                    )


@cli.cli_plugin(
    help=(
        'Change the current prefix link, so the default prefix that is used '
        'is a new one'
    ),
)
@cli.cli_plugin_add_argument(
    'prefix_name',
    action='store',
    help='Name of the prefix to set as current',
)
@utils.in_prefix(
    prefix_class=prefix.Prefix,
    workdir_class=Workdir,
)
def set_current(prefix_name, parent_workdir, **kwargs):
    """
    Changes the current to point to the given prefix

    Args:
        prefix_name(str): name of the prefix to set the current to
        workdir(str): path to the workdir to change the current of
    """
    parent_workdir.set_current(new_current=prefix_name)



#
# Copyright 2016 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
import os
import re
from collections import defaultdict
from io import StringIO
from warnings import warn

import configparser
from xdg import BaseDirectory as base_dirs

from lago.constants import CONFS_PATH, CONFIG_DEFAULTS
from lago.utils import argparse_to_ini


def _get_configs_path():
    """Get a list of possible configuration files, from the following
    sources:
    1. All files that exists in constants.CONFS_PATH.
    2. All XDG standard config files for "lago.conf", in reversed
    order of importance.


    Returns:
        list(str): list of files

    """

    paths = []
    xdg_paths = [
        path for path in base_dirs.load_config_paths('lago', 'lago.conf')
    ]
    paths.extend([path for path in CONFS_PATH if os.path.exists(path)])
    paths.extend(reversed(xdg_paths))
    return paths


def get_env_dict(root_section):
    """Read all Lago variables from the environment.
    The lookup format is:
    LAGO_VARNAME - will land into 'lago' section
    LAGO__SECTION1__VARNAME - will land into 'section1' section, notice
    the double '__'.
    LAGO__LONG_SECTION_NAME__VARNAME - will land into 'long_section_name'


    Returns:
        dict: dict of section configuration dicts

    Examples:
        >>> os.environ['LAGO_GLOBAL_VAR'] = 'global'
        >>> os.environ['LAGO__INIT__REPO_PATH'] = '/tmp/store'
        >>>
        >>> config.get_env_dict()
        {'init': {'repo_path': '/tmp/store'}, 'lago': {'global_var': 'global'}}

    """
    env_lago = defaultdict(dict)
    decider = re.compile(
        (
            r'^{0}(?:_(?!_)|(?P<has>__))'
            r'(?(has)(?P<section>.+?)__)'
            r'(?P<name>.+)$'
        ).format(root_section.upper())
    )

    for key, value in os.environ.iteritems():
        match = decider.match(key)
        if not match:
            continue
        if not match.group('name') or not value:
            warn(
                'empty environment variable definition:'
                '{0}, ignoring.'.format(key)
            )
        else:
            section = match.group('section') or root_section
            env_lago[section.lower()][match.group('name').lower()] = value
    return dict(env_lago)


class ConfigLoad(object):
    """Merges configuration parameters from 3 different sources:
    1. Enviornment vairables
    2. config files in .INI format
    3. argparse.ArgumentParser

    The assumed order(but not necessary) order of calls is:
    load() - load from config files and environment variables
    update_parser(parser) - update from the declared argparse parser
    update_args(args) - update from passed arguments to the parser

    """

    def __init__(self, root_section='lago', defaults={}):
        """__init__
        Args:
            root_section (str): root section in the init
            defaults (dict): Default dictonary to load, can be empty.
        """

        self.root_section = root_section
        self._defaults = defaults
        self._config = defaultdict(dict)
        self._config.update(self.load())
        self._parser = None

    def load(self):
        """
        Load all configurations from available resources, skip if empty:

            1. :attr:`default`` dict passed to :func:`ConfigLoad.__init__`.
            2. Custom paths as defined in :attr:`CONFS_PATH` in
                :class:`~lago.constants`.
            3. XDG standard paths.
            4. Environment variables.

        Returns:
            dict: dict of dicts.

        """

        configp = configparser.ConfigParser()
        configp.read_dict(self._defaults)
        for path in _get_configs_path():
            try:
                with open(path, 'r') as config_file:
                    configp.read_file(config_file)
            except IOError:
                pass
        configp.read_dict(get_env_dict(self.root_section))
        return {s: dict(configp.items(s)) for s in configp.sections()}

    def update_args(self, args):
        """Update config dictionary with parsed args, as resolved by argparse.
        Only root positional arguments that already exist will overridden.

        Args:
            args (namespace): args parsed by argparse

        """

        for arg in vars(args):
            if self.get(arg) and getattr(args, arg) is not None:
                self._config[self.root_section][arg] = getattr(args, arg)

    def update_parser(self, parser):
        """Update config dictionary with declared arguments in an argparse.parser
        New variables will be created, and existing ones overridden.

        Args:
            parser (argparse.ArgumentParser): parser to read variables from

        """

        self._parser = parser
        ini_str = argparse_to_ini(parser)
        configp = configparser.ConfigParser(allow_no_value=True)
        configp.read_dict(self._config)
        configp.read_string(ini_str)
        self._config.update(
            {s: dict(configp.items(s))
             for s in configp.sections()}
        )

    def get(self, *args):
        """Get a variable from the default section
        Args:
            *args (args): dict.get() args

        Returns:
            str: config variable

        """

        return self._config[self.root_section].get(*args)

    def __getitem__(self, key):
        """Get a variable from the default section, good for fail-fast
        if key does not exists.

        Args:
            key (str): key

        Returns:
            str: config variable

        """

        return self._config[self.root_section][key]

    def get_section(self, *args):
        """get a section dictionary
        Args:

        Returns:
            dict: section config dictionary

        """

        return self._config.get(*args)

    def get_ini(self, incl_unset=False):
        """Return the config dictionary in INI format
        Args:
            incl_unset (bool): include variables with no defaults.

        Returns:
            str: string of the config file in INI format

        """

        configp = configparser.ConfigParser(allow_no_value=True)
        configp.read_dict(self._config)
        with StringIO() as config_ini:
            if self._parser:
                self._parser.set_defaults(
                    **self.get_section(self.root_section)
                )
                argparse_ini = argparse_to_ini(
                    parser=self._parser, incl_unset=incl_unset
                )
                return argparse_ini
            else:
                configp.write(config_ini)
                return config_ini.getvalue()

    def __repr__(self):
        return self._config.__repr__()

    def __str__(self):
        return self._config.__str__()


config = ConfigLoad(defaults=CONFIG_DEFAULTS)



#
# Copyright 2016-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#

LIBEXEC_DIR = '/usr/libexec/lago/'
"""
LIBEXEC_DIR -
"""

CONFS_PATH = ['/etc/lago/lago.conf']
"""
CONFS_PATH - default path to first look for configuration files.
"""

CONFIG_DEFAULTS = {
    'lago':
        {
            'logdepth': 3,
            'loglevel': 'info',
            'ssh_user': 'root',
            'ssh_password': '123456',
            'ssh_tries': 100,
            'ssh_timeout': 10,
            'libvirt_url': 'qemu:///system',
            'default_vm_type': 'default',
            'default_vm_provider': 'local-libvirt',
            'lease_dir': '/var/lib/lago/subnets',
            'prefix_name': 'current',
            'reposync_dir': '/var/lib/lago',
        },
    'init':
        {
            'template_repo_path':
                'http://templates.ovirt.org/repo/repo.metadata',
            'template_store':
                '/var/lib/lago/store',
            'template_repos':
                '/var/lib/lago/repos',
        }
}



#!/usr/bin/env python2
#
# Copyright 2014 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
"""
"""
from stevedore import ExtensionManager
import logging

LOGGER = logging.getLogger(__name__)

#: Map of plugin type string -> setuptools entry point
PLUGIN_ENTRY_POINTS = {
    'cli': 'lago.plugins.cli',
    'out': 'lago.plugins.output',
    'vm': 'lago.plugins.vm',
    'vm-service': 'lago.plugins.vm_service',
    'vm-provider': 'lago.plugins.vm_provider',
}


class PluginError(Exception):
    pass


class NoSuchPluginError(PluginError):
    pass


class Plugin(object):
    """
    Base class for all the plugins
    """
    pass


def load_plugins(namespace, instantiate=True):
    """
    Loads all the plugins for the given namespace

    Args:
        namespace(str): Namespace string, as in the setuptools entry_points
        instantiate(bool): If true, will instantiate the plugins too

    Returns:
        dict of str, object: Returns the list of loaded plugins
    """
    mgr = ExtensionManager(
        namespace=namespace,
        on_load_failure_callback=(
            lambda _, ep, err: LOGGER.
            warning('Could not load plugin {}: {}'.format(ep.name, err))
        )
    )
    if instantiate:
        plugins = dict(
            (
                ext.name,
                ext.plugin if isinstance(ext.plugin, Plugin) else ext.plugin()
            ) for ext in mgr
        )
    else:
        plugins = dict((ext.name, ext.plugin) for ext in mgr)

    return plugins



#
# Copyright 2015-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
"""
This module defines the special logging tools that lago uses
"""
from future.builtins import super
import logging
import logging.config
import os
import re
import traceback
import datetime
import threading
import uuid as uuid_m
from collections import (
    OrderedDict,
    deque,
)
from functools import wraps

#: Message to be shown when a task is started
START_TASK_MSG = ''
#: Message template that will trigger a task
START_TASK_TRIGGER_MSG = 'start task%s'
#: Regexp that will match the above template
START_TASK_REG = re.compile('start task(?P<task_name>.*)')
#: Message to be shown when a task is ended
END_TASK_MSG = 'Success'
END_TASK_ON_ERROR_MSG = 'ERROR'
#: Message template that will trigger a task end
END_TASK_TRIGGER_MSG = 'end task%s'
#: Regexp that will match the above template
END_TASK_REG = re.compile('end task(?P<task_name>.*)')
#: Message template that will always shoud the messago
ALWAYS_SHOW_TRIGGER_MSG = 'force-show:%s'
#: Regexp that will match the above template
ALWAYS_SHOW_REG = re.compile('force-show:(?P<message>.*)')


class ColorFormatter(logging.Formatter):
    """
    Formatter to add colors to log records
    """
    DEFAULT = '\x1b[0m'
    RED = '\x1b[31m'
    GREEN = '\x1b[32m'
    YELLOW = '\x1b[33m'
    CYAN = '\x1b[36m'
    WHITE = '\x1b[37m'
    NONE = ''
    CRITICAL = RED
    ERROR = RED
    WARNING = YELLOW
    INFO = CYAN
    DEBUG = NONE

    @classmethod
    def colored(cls, color, message):
        """
        Small function to wrap a string around a color

        Args:
            color (str): name of the color to wrap the string with, must be one
                of the class properties
            message (str): String to wrap with the color

        Returns:
            str: the colored string
        """
        return getattr(cls, color.upper()) + message + cls.DEFAULT

    def format(self, record):
        """
        Adds colors to a log record and formats it with the default

        Args:
            record (logging.LogRecord): log record to format

        Returns:
            str: The colored and formatted record string
        """
        level = record.levelno

        if level >= logging.CRITICAL:
            color = self.CRITICAL
        elif level >= logging.ERROR:
            color = self.ERROR
        elif level >= logging.WARNING:
            color = self.WARNING
        elif level >= logging.INFO:
            color = self.INFO
        elif level >= logging.DEBUG:
            color = self.DEBUG
        else:
            color = self.DEFAULT

        message = super().format(record)
        if record.args:
            try:
                message = message % record.args
            except TypeError:
                # this happens when the message itself has some %s symbols, as
                # in traces for tracedumps
                pass

        return color + message + self.DEFAULT


class Task(deque):
    """
    Small wrapper around deque to add the failed status and name to a task

    Attributes:
        name (str): name for this task
        failed (bool): If this task has failed or not (if there was any error
            log shown during it's execution)
        force_show (bool): If set, will show any log records generated inside
            this task even if it's out of nested depth limit
    """

    def __init__(self, name, *args, **kwargs):
        """
        Args:
            name (str): name for this task
            *args: any :class:`deque` args
            *kwargs: any :class:`deque` kwargs
        """
        self.failed = False
        self.force_show = False
        self.name = name
        self.start_time = datetime.datetime.now()
        super(deque, self).__init__(*args, **kwargs)

    def __str__(self):
        return (
            '%s(failed=%s, force_show=%s, len=%d)' %
            (self.name, self.failed, self.force_show, len(self))
        )

    def elapsed_time(self):
        return str(datetime.datetime.now() - self.start_time).rsplit('.', 1)[0]


class ContextLock(object):
    """
    Context manager to thread lock a block of code
    """

    def __init__(self):
        self.lock = threading.Lock()

    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()


class TaskHandler(logging.StreamHandler):
    """
    This log handler will use the concept of tasks, to hide logs, and will show
    all the logs for the current task if there's a logged error while running
    that task.

    It will hide any logs that belong to nested tasks that have more than
    ``task_tree_depth`` parent levels, and for the ones that are above that
    level, it will show only the logs that have a loglevel above
    ``level``.

    You can force showing a log record immediately if you use the
    :func:`log_always` function bypassing all the filters.

    If there's a log record with log level higher than ``dump_level`` it will
    be considered a failure, and all the logs for the current task that have a
    log level above ``level`` will be shown no matter at which depth the task
    belongs to. Also, all the parent tasks will be tagged as error.

    Attributes:
        formatter (logging.LogFormatter): formatter to use
        initial_depth (int): Initial depth to account for, in case this handler
            was created in a subtask
        tasks_by_thread (dict of str: OrderedDict of str: Task): List of
            thread names, and their currently open tasks with  their latest
            log records
        dump_level (int): log level from which to consider a log record as
            error
        buffer_size (int): Size of the log record deque for each task, the
            bigger, the more records it can show in case of error but the more
            memory it will use
        task_tree_depth (int): number of the nested level to show start/end
            task logs for, if -1 will show always
        level (int): Log level to show logs from if the depth limit is not
            reached
        main_failed (bool): used to flag from a child thread that the main
            should fail any current task
        _tasks_lock (ContextLock): Lock for the tasks_by_thread dict
        _main_thread_lock (ContextLock): Lock for the main_failed bool
    """
    #: List of chars to show as task prefix, to ease distinguishing them
    TASK_INDICATORS = ['@', '#', '*', '-', '~']

    def __init__(
        self,
        initial_depth=0,
        task_tree_depth=-1,
        buffer_size=2000,
        dump_level=logging.ERROR,
        level=logging.NOTSET,
        formatter=ColorFormatter,
    ):
        super().__init__()
        self.formatter = formatter
        self.initial_depth = initial_depth
        self.tasks_by_thread = {}
        self.dump_level = dump_level
        self.buffer_size = buffer_size
        self.task_tree_depth = task_tree_depth
        self.level = level
        self.main_failed = False
        self._tasks_lock = ContextLock()
        self._main_thread_lock = ContextLock()

    @property
    def cur_task(self):
        """
        Returns:
            str: the current active task
        """
        return self.tasks.keys()[-1] if self.tasks else None

    @property
    def cur_thread(self):
        """
        Returns:
            str: Name of the current thread
        """
        return threading.current_thread().name

    @property
    def tasks(self):
        """
        Returns:
            OrderedDict of str, Task: list of task names and log records for
                each for the current thread
        """
        return self.get_tasks(thread_name=self.cur_thread)

    def get_tasks(self, thread_name):
        """
        Args:
            thread_name (str): name of the thread to get the tasks for

        Returns:
            OrderedDict of str, Task: list of task names and log records for
                each for the given thread
        """
        if thread_name not in self.tasks_by_thread:
            with self._tasks_lock:
                self.tasks_by_thread[thread_name] = OrderedDict()

        return self.tasks_by_thread[thread_name]

    @property
    def cur_depth_level(self):
        """
        Returns:
            int: depth level for the current task
        """
        cur_level = self.initial_depth
        if not self.am_i_main_thread:
            cur_level += len(self.get_tasks(thread_name='MainThread'))

        return cur_level + len(self.tasks)

    @property
    def am_i_main_thread(self):
        """
        Returns:
            bool: if the current thread is the main thread
        """
        return threading.current_thread().name == 'MainThread'

    def mark_main_tasks_as_failed(self):
        """
        Flags to the main thread that all it's tasks sholud fail

        Returns:
            None
        """
        if self.am_i_main_thread:
            return

        with self._main_thread_lock:
            self.main_failed = True

    def should_show_by_depth(self, cur_level=None):
        """
        Args:
            cur_level (int): depth level to take into account

        Returns:
            bool: True if the given depth level should show messages (not
                taking into account the log level)
        """
        if cur_level is None:
            cur_level = self.cur_depth_level

        return (self.task_tree_depth < 0 or self.task_tree_depth >= cur_level)

    def should_show_by_level(self, record_level, base_level=None):
        """
        Args:
            record_level (int): log level of the record to check
            base_level (int or None): log level to check against, will use the
                object's :class:`dump_level` if None is passed

        Returns:
            bool: True if the given log record should be shown according to the
                log level
        """
        if base_level is None:
            base_level = self.dump_level

        return record_level >= base_level

    def handle_new_task(self, task_name, record):
        """
        Do everything needed when a task is starting

        Params:
            task_name (str): name of the task that is starting
            record (logging.LogRecord): log record with all the info

        Returns:
            None
        """
        record.msg = ColorFormatter.colored('default', START_TASK_MSG)
        record.task = task_name

        self.tasks[task_name] = Task(name=task_name, maxlen=self.buffer_size)
        if self.should_show_by_depth():
            self.pretty_emit(record, is_header=True)

    def mark_parent_tasks_as_failed(self, task_name, flush_logs=False):
        """
        Marks all the parent tasks as failed

        Args:
            task_name (str): Name of the child task
            flush_logs (bool): If ``True`` will discard all the logs form
                parent tasks

        Returns:
            None
        """
        for existing_task_name in self.tasks:
            if existing_task_name == task_name:
                break

            if flush_logs:
                self.tasks[existing_task_name].clear()

            self.tasks[existing_task_name].failed = True

        self.mark_main_tasks_as_failed()

    def close_children_tasks(self, parent_task_name):
        """
        Closes all the children tasks that were open

        Args:
            parent_task_name (str): Name of the parent task

        Returns:
            None
        """
        if parent_task_name not in self.tasks:
            return

        while self.tasks:
            next_task = reversed(self.tasks.keys()).next()
            if next_task == parent_task_name:
                break
            del self.tasks[next_task]

    def handle_closed_task(self, task_name, record):
        """
        Do everything needed when a task is closed

        Params:
            task_name (str): name of the task that is finishing
            record (logging.LogRecord): log record with all the info

        Returns:
            None
        """
        if task_name not in self.tasks:
            return

        if self.main_failed:
            self.mark_parent_tasks_as_failed(self.cur_task)

        if self.tasks[task_name].failed:
            record.msg = ColorFormatter.colored('red', END_TASK_ON_ERROR_MSG)
        else:
            record.msg = ColorFormatter.colored('green', END_TASK_MSG)

        record.msg += ' (in %s)' % self.tasks[task_name].elapsed_time()

        if self.should_show_by_depth() or self.tasks[task_name].force_show:
            if self.tasks[task_name].force_show:
                self.handle_error()

            self.pretty_emit(record, is_header=True)

        self.close_children_tasks(task_name)
        self.tasks.pop(task_name)

    def handle_error(self):
        """
        Handles an error log record that should be shown

        Returns:
            None
        """
        if not self.tasks:
            return

        # All the parents inherit the failure
        self.mark_parent_tasks_as_failed(
            self.cur_task,
            flush_logs=True,
        )

        # Show the start headers for all the parent tasks if they were not
        # shown by the depth level limit
        for index, task in enumerate(self.tasks.values()):
            if self.should_show_by_depth(index + 1):
                continue

            start_task_header = logging.LogRecord(
                '', logging.INFO, '', 0, '', [], None
            )
            start_task_header.msg = ColorFormatter.colored(
                'default',
                START_TASK_MSG,
            )
            start_task_header.task = task.name
            self.pretty_emit(
                start_task_header,
                is_header=True,
                task_level=index + 1,
            )

        # Show now all the cached logs for the current task
        for old_record in self.tasks[self.cur_task]:
            self.pretty_emit(old_record)

        self.tasks[self.cur_task].clear()

    def get_task_indicator(self, task_level=None):
        """
        Args:
            task_level (int or None): task depth level to get the indicator
                for, if None, will use the current tasks depth

        Returns:
            str: char to prepend to the task logs to indicate it's level
        """
        if task_level is None:
            task_level = len(self.tasks)
        return self.TASK_INDICATORS[task_level % len(self.TASK_INDICATORS)]

    def pretty_emit(self, record, is_header=False, task_level=None):
        """
        Wrapper around the :class:`logging.StreamHandler` emit method to add
        some decoration stuff to the message

        Args:
            record (logging.LogRecord): log record to emit
            is_header (bool): if this record is a header, usually, a start or
                end task message
            task_level (int): If passed, will take that as the current nested
                task level instead of calculating it from the current tasks

        Returns:
            None
        """
        task = record.task or self.cur_task
        if task_level is None:
            task_level = self.cur_depth_level

        if is_header:
            extra_prefix = (
                self.get_task_indicator(task_level - 1) + ' ' +
                ('' if self.am_i_main_thread else '[%s] ' % self.cur_thread) +
                task + ': '
            )
            record.levelno = logging.INFO
        else:
            extra_prefix = '  ' + self.get_task_indicator(task_level) + ' '

        if task:
            record.msg = (
                '  ' * (task_level - 1) + extra_prefix + str(record.msg)
            )

        super().emit(record)
        super().flush()

    def emit(self, record):
        """
        Handle the given record, this is the entry point from the python
        logging facility

        Params:
            record (logging.LogRecord): log record to handle

        Returns:
            None
        """
        record.task = self.cur_task

        if record.levelno >= self.dump_level and self.cur_task:
            self.tasks[self.cur_task].failed = True
            self.tasks[self.cur_task].force_show = True

        # Makes no sense to start a task with an error log
        is_start = START_TASK_REG.match(str(record.msg))
        if is_start:
            self.handle_new_task(is_start.groupdict()['task_name'], record)
            return

        is_end = END_TASK_REG.match(str(record.msg))
        if is_end:
            self.handle_closed_task(is_end.groupdict()['task_name'], record)
            return

        force_show_record = ALWAYS_SHOW_REG.match(str(record.msg))
        if force_show_record:
            record.msg = force_show_record.groupdict()['message']
            self.pretty_emit(record)

        if (
            not force_show_record and self.should_show_by_level(record)
            and self.should_show_by_depth()
        ):
            self.pretty_emit(record)
            return

        if self.cur_task:
            self.tasks[self.cur_task].append(record)


class LogTask(object):
    """
    Context manager for a log task

    Example:
        >>> with LogTask('mytask'):
        ...     pass
    """

    def __init__(
        self,
        task,
        logger=logging,
        level='info',
        propagate_fail=True,
        uuid=None,
    ):
        self.task = task
        self.logger = logger
        self.level = level
        self.propagate = propagate_fail
        if uuid is None:
            self.uuid = uuid_m.uuid4()
        self.header = self.task
        if self.level != 'info':
            self.header = ':{0}:{1}:'.format(str(self.uuid), self.task)

    def __enter__(self):
        getattr(self.logger, self.level)(START_TASK_TRIGGER_MSG % self.header)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type and self.propagate:
            end_log_task(self.header, level='error')
            str_tb = ''.join(traceback.format_tb(exc_tb))
            self.logger.debug(str_tb)
            return False
        else:
            getattr(self.logger,
                    self.level)(END_TASK_TRIGGER_MSG % self.header)


def log_task(
    task, logger=logging, level='info', propagate_fail=True, uuid=None
):
    """
    Parameterized decorator to wrap a function in a log task

    Example:
        >>> @log_task('mytask')
        ... def do_something():
        ...     pass
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with LogTask(
                task,
                logger=logger,
                level=level,
                propagate_fail=propagate_fail,
                uuid=uuid
            ):
                return func(*args, **kwargs)

        return wrapper

    return decorator


def start_log_task(task, logger=logging, level='info'):
    """
    Starts a log task

    Args:
        task (str): name of the log task to start
        logger (logging.Logger): logger to use
        level (str): log level to use

    Returns:
        None
    """
    getattr(logger, level)(START_TASK_TRIGGER_MSG % task)


def end_log_task(task, logger=logging, level='info'):
    """
    Ends a log task

    Args:
        task (str): name of the log task to end
        logger (logging.Logger): logger to use
        level (str): log level to use

    Returns:
        None
    """
    getattr(logger, level)(END_TASK_TRIGGER_MSG % task)


def log_always(message):
    """
    Wraps the given message with a tag that will make it be always logged by
    the task logger

    Args:
        message (str): message to wrap with the tag

    Returns:
        str: tagged message that will get it shown immediately by the task
            logger
    """
    return ALWAYS_SHOW_TRIGGER_MSG % message


def hide_paramiko_logs():
    paramiko_logger = logging.getLogger('paramiko.transport')
    paramiko_logger.propagate = False
    paramiko_logger.setLevel(logging.ERROR)


def hide_stevedore_logs():
    """
    Hides the logs of stevedore, this function was
    added in order to support older versions of stevedore

    We are using the NullHandler in order to get rid from
    'No handlers could be found for logger...' msg

    Returns:
        None
    """
    stevedore_logger = logging.getLogger('stevedore.extension')
    stevedore_logger.propagate = False
    stevedore_logger.setLevel(logging.ERROR)
    stevedore_logger.addHandler(logging.NullHandler())


def setup_prefix_logging(logdir):
    """
    Sets up a file logger that will create a log in the given logdir (usually a
    lago prefix)

    Args:
        logdir (str): path to create the log into, will be created if it does
            not exist

    Returns:
        None
    """
    if not os.path.exists(logdir):
        os.mkdir(logdir)

    file_handler = logging.FileHandler(
        filename=os.path.join(logdir, 'lago.log'),
    )
    file_formatter = get_default_log_formatter()
    file_handler.setFormatter(file_formatter)
    logging.root.addHandler(file_handler)
    hide_paramiko_logs()
    hide_stevedore_logs()


def get_default_log_formatter():
    return logging.Formatter(
        fmt=(
            '%(asctime)s::%(filename)s::%(funcName)s::%(lineno)s::'
            '%(name)s::%(levelname)s::%(message)s'
        ),
    )
